import time
import logging


from lxf.settings import get_logging_level

logger = logging.getLogger('Measures')
fh = logging.FileHandler('./logs/Measures.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)

def measure_time(func) :
    def inner(*arg,**kwargs):
        t=time.time()
        result = func(*arg,**kwargs)
        logger.debug(f"{func.__name__} executed in {(time.time()-t):.4f} seconds ")
        return result
    return inner

def  measure_time_async(func) :
    async def inner(*arg,**kwargs):
        t=time.time()
        result = await func(*arg,**kwargs)
        logger.debug(f"{func.__name__} executed in {(time.time()-t):.4f} seconds ")
        return result
    return inner





import os

import logging
import re


from lxf.settings import get_logging_level

logger = logging.getLogger('cni recognize')
fh = logging.FileHandler('./logs/cni_recognize.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)

from typing import Tuple
import cv2
from cv2.typing import MatLike


from lxf.domain.cni import Cni
from lxf.domain.cni_zones import CniDetectedZone
from lxf.services.measure_time import measure_time
from lxf.ai.ocr.cni_utilities.detection.cni_detection import detection , detectFront, detectBack
from lxf.ai.ocr.text_extractor import extract_text_from_image


def extract_id_number(text: str) -> str:
    """
    Extrait le numéro d'identité à partir du texte OCR.
    """
    match = re.search(r'\b\d{8,}\b', text)

    return match.group(0) if match else None

def extract_personal_info(text: str) -> Tuple[str, str]:
    """
    Extrait le nom, le prénom à partir du texte OCR.
    """
    # Cherche le prénom et le nom
    first_name = None
    last_name = None
    match_name = re.search(r"Prénom(?:\(s\))?\s*:?\s*(.+)", text, re.IGNORECASE)
    if match_name:
        first_name = match_name.group(1).strip()

    match_last_name = re.search(r"Nom\s*:?\s*([A-ZÉÈÀ'\-\.]+(?:\sEpouse\s[A-ZÉÈÀ'\-\.]+)?)", text, re.IGNORECASE)
    if match_last_name:
        last_name = match_last_name.group(1).strip()

    return first_name, last_name


def extract_birth_date(text:str)->str:
    """
    Extrait la date de naissance a partir du texte OCR
    """
    match_date = re.search(
    r"N[ée]?\(?e?\)?\s*le[:;]?\s*(\d{2}[\.\s]\d{2}[\.\s]?\d{4})", 
    text, 
    re.IGNORECASE
)
    if match_date:
        return match_date.group(1).replace(" ", ".")
    return None

def extract_dates(text: str) -> Tuple[str, str]:
    """
    Extrait deux dates d'un texte et les attribue :
    - La première date trouvée est "valable jusqu'à".
    - La deuxième date trouvée est "délivré le".
    """
    dates = re.findall(r"\d{2}\s*[.,]\s*\d{2}\s*[.,]?\s*\d{4}", text)
    
    standardized_dates = [date.replace(" ", ".").replace(",", ".") for date in dates]
    
    expiry_date = standardized_dates[0] if len(standardized_dates) > 0 else None
    issue_date = standardized_dates[1] if len(standardized_dates) > 1 else None

    return expiry_date, issue_date


def extract_address(text: str) -> str:
    """
    Extrait l'adresse à partir d'un texte OCR.
    
    """
    match = re.search(
        r"[Aa]dresse\s*:?\s*(.+?)\s*(Carte valable jusqu'au|délivrée le)", 
        text
    )
    if match:
        return match.group(1).strip()
    return None

@measure_time
def recognize(filename: str) -> Cni:
    """
        Applique l'OCR sur une image de CNI pour extraire ses informations.
        return: Cni
    """
    if not os.path.exists(filename):
        msg = f"Le fichier {filename} n'existe pas"
        logger.error(msg)
        raise Exception(msg)
    
    result: CniDetectedZone = detection(filename)
    if result is None:
        logger.error(f"Aucune zone détectée ({filename})")
        return None
    
    cni = Cni(is_recto=result.is_recto)


    if result.is_recto:
        # Traitement du recto
        zones = detectFront(result)
        if zones is None:
            logger.error(f"Aucune zone détectée pour le recto ! {filename}")
            return None
        
        # Zone ID
        source = zones.id
        id_image = process_image(source)

        if id_image.any:
            id_text = extract_text_from_image(id_image)
            cni.id_number = extract_id_number(id_text)
        
        # Zone Informations (nom, prénom, etc.)
        source = zones.info
        info_image = process_image(source)

        if info_image.any:
            info_text = extract_text_from_image(info_image)
            cni.first_name, cni.last_name= extract_personal_info(info_text)
        
        # Zone MRZ
        source = zones.mrz
        mrz_image = process_image(source)

        if mrz_image.any:
            mrz_text = extract_text_from_image(mrz_image,use_tesseract=False)
            cni.mrz= mrz_text.strip().replace(" ","")
        

        #zone date naissance
        source = zones.dat_of_birth
        date_image=process_image(source)
        if date_image.any:
            date_text = extract_text_from_image(date_image)
            date_text=extract_birth_date(date_text)
            cni.date_of_birth=date_text

    else:
        # Traitement du verso
        info = detectBack(result.image)
        verso = process_image_verso(info)

        if verso.any:
            verso_text = extract_text_from_image(verso,
                                                 use_tesseract=False)
            expiry_date, issue_date=extract_dates(verso_text)
            cni.expiry_date=expiry_date
            cni.issue_date=issue_date
            cni.address=extract_address(verso_text)
            
    
    return cni

@measure_time
def process_image_verso(source)->MatLike:
    """
    Prétraite le verso de la cni pour l'OCR.
    :param img: Image source.
    :return: Image prétraitée.
    """
    if source is None or source.size == 0:
        raise Exception(f"L'image source {source} est vide ou non initialisée.")

    logger.debug(f"Dimensions de l'image avant traitement : {source.shape}")

    gray: MatLike = cv2.cvtColor(source, cv2.COLOR_BGR2GRAY)
    if gray is None or gray.size == 0:
        raise Exception("Erreur lors de la conversion en niveaux de gris.")

    final:MatLike=cv2.bilateralFilter(gray, 11,75,75)
    thresh = cv2.adaptiveThreshold(
        final, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 11, 2)

    return thresh


@measure_time
def apply_truncate_filter(img: MatLike) -> MatLike:
    """
    Applique un filtre Truncate pour réduire le bruit.
    :param img: Image source.
    :return: Image filtrée.
    """
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    
    _, truncate_img = cv2.threshold(gray, 140, 255, cv2.THRESH_TRUNC)
    
    return truncate_img


def process_image(img: MatLike) -> MatLike:

    """
    Prétraite une image pour l'OCR.
    :param img: Image source.
    :return: Image prétraitée.
    """
    if img is None or img.size == 0:
        raise Exception(f"L'image source {img}est vide ou non initialisée.")
    
    truncate_img:MatLike = apply_truncate_filter(img)
    
    blurred = cv2.GaussianBlur(truncate_img, (5, 5), 0)

    thresh = cv2.adaptiveThreshold(
        truncate_img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 11, 2
    )
    
    return thresh





import logging
from typing import Tuple
from lxf.settings import get_logging_level
import mediapipe as mp

#logger
logger = logging.getLogger('rois cni')
fh = logging.FileHandler('./logs/cni_rois.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)

import cv2
from cv2.typing import MatLike , RotatedRect
import pytesseract
from pdf2image import convert_from_path, convert_from_bytes
from pdf2image.exceptions import (
    PDFInfoNotInstalledError,
    PDFPageCountError,
    PDFSyntaxError
)
import numpy as np
import re
import os
import locale
from enum import IntEnum


# Credits to https://becominghuman.ai/how-to-automatically-deskew-straighten-a-text-image-using-opencv-a0c30aed83df
def getSkewAngle(cvImage) -> float:
    # Prep image, copy, convert to gray scale, blur, and threshold
    newImage = cvImage.copy()
    
    gray = cv2.cvtColor(newImage, cv2.COLOR_BGR2GRAY)

    thresh, img_bw = cv2.threshold(gray,150,250,cv2.THRESH_BINARY)
    #cv2.imwrite("./data/images//temp/deskew_img_bw.png", img_bw)

    no_noise = noise_removal(img_bw)
    #cv2.imwrite("./data/images/temp/deskwe_no_noise.jpg", no_noise)

    blur = cv2.GaussianBlur(no_noise, (9, 9), 0)
    #cv2.imwrite("./data/images/temp/deskew_blur.jpg",blur)    

    thresh = cv2.threshold(blur, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]
    #cv2.imwrite("./data/images/temp/deskew_thresh.jpg",thresh)
    # Apply dilate to merge text into meaningful lines/paragraphs.
    # Use larger kernel on X axis to merge characters into single line, cancelling out any spaces.
    # But use smaller kernel on Y axis to separate between different blocks of text
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (15, 9))
    dilate = cv2.dilate(thresh, kernel, iterations=1)
    #cv2.imwrite("./data/images/temp/deskew_dilate.jpg",dilate)
    
    # Find all contours
    contours, hierarchy = cv2.findContours(dilate, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
    contours = sorted(contours, key = cv2.contourArea, reverse = True)
    for c in contours:
        rect = cv2.boundingRect(c)
        x,y,w,h = rect
        cv2.rectangle(newImage,(x,y),(x+w,y+h),(0,255,0),2)

    # Find largest contour and surround in min area box
    largestContour = contours[0]
    #print (len(contours))
    minAreaRect = cv2.minAreaRect(largestContour)
    box = np.intp(cv2.boxPoints(minAreaRect))
    cv2.drawContours(newImage, [box], 0, (0,0,255), 3)
    #cv2.imwrite("./data/images/temp/boxes.jpg", newImage)
    # Determine the angle. Convert it to the value that was originally used to obtain skewed image
    angle = minAreaRect[-1]
    #print(f"Deskew angle:{angle}")
    if angle < -45:
        angle = 90 + angle
    return -1.0 * angle



def correct_orientation(cvImage : MatLike, rect: RotatedRect)->MatLike:
    # Calculate the angle from rect
    angle : float= rect[2]
    if rect[1][0] < rect[1][1]:
        angle = 90 - rect[2]
    else:
        angle = rect[2]
    
    center_rect = rect[0]
    (h1, w1) = cvImage.shape[:2]
    M : MatLike = cv2.getRotationMatrix2D(center_rect, angle, 1.0)
    cos:float = np.abs(M[0, 0])
    sin:float = np.abs(M[0, 1])
    new_w:int = int((w1 * cos) + (h1 * sin))
    new_h:int = int((h1 * cos) + (w1 * sin))
    M[0, 2] += (new_w / 2) - center_rect[0]
    M[1, 2] += (new_h / 2) - center_rect[1]
    deskewed_image:MatLike = cv2.warpAffine(cvImage, M, (new_w, new_h))
    
    return deskewed_image
    
# Rotate the image around its center
def rotateImage(cvImage, angle:float):
    """
    Effectue une rotation de l'image autour de son centre.
    """
    (h, w) = cvImage.shape[:2]
    center = (w // 2, h // 2)
    # Créer la matrice de rotation
    M = cv2.getRotationMatrix2D(center, angle, 1.0)
    # Appliquer la transformation affine
    rotated = cv2.warpAffine(cvImage, M, (w, h), flags=cv2.INTER_CUBIC, borderMode=cv2.BORDER_REPLICATE)
    return rotated

def to_gray(image):
    gray = cv2.cvtColor(image,cv2.COLOR_BGR2GRAY)
    return gray

def noise_removal(image):
    import numpy as np
    kernel = np.ones((1, 1), np.uint8)
    image = cv2.dilate(image, kernel, iterations=1)
    kernel = np.ones((1, 1), np.uint8)
    image = cv2.erode(image, kernel, iterations=1)
    image = cv2.morphologyEx(image, cv2.MORPH_CLOSE, kernel)
    image = cv2.medianBlur(image, 3)
    return (image)

def thin_font(image):
    import numpy as np
    image = cv2.bitwise_not(image)
    kernel = np.ones((1,1),np.uint8)
    image = cv2.erode(image, kernel, iterations=1)
    image = cv2.bitwise_not(image)
    return (image)

def remove_borders(image):
    contours, hiearchy = cv2.findContours(image, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cntsSorted = sorted(contours, key=lambda x:cv2.contourArea(x))
    cnt = cntsSorted[-1]
    x, y, w, h = cv2.boundingRect(cnt)
    crop = image[y:y+h, x:x+w]
    return (crop)

def sortKeyByColumns(x, hmax,hmoy,wmoy):    
    x,y,w,h = cv2.boundingRect(x)
    col = x//wmoy
    line = y//hmoy
    return x+col*hmax+line*hmoy

def sortKeyByRows(x, hmax,hmoy,wmoy):
    x,y,w,h = cv2.boundingRect(x)
    col = x//wmoy
    line = y//hmoy
    return x+col*wmoy+line*hmoy


class ReadRirection(IntEnum):
    row=0
    col=1

def detect_struct(image, name:str="X",
                  max_high:float=1900.0,
                  min_high:float=80.0,
                  max_wide:float=1500.00,
                  min_wide:float=60.00,
                  direction:ReadRirection=ReadRirection.row,
                  displayIntermediare:bool=False) :
    
    """
    Compute the ROI of the image 
    """
    
    # convert to gray
    gray = cv2.cvtColor(image,cv2.COLOR_BGR2GRAY)
    blur = cv2.GaussianBlur(gray,(17,19),0)
    
    thresh=cv2.threshold(blur,150,250,cv2.THRESH_BINARY_INV+cv2.THRESH_OTSU)[1]
 
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT,(21,21))
   
    dilate = cv2.dilate(thresh, kernel, iterations=1)

    cv2.imwrite("./ocr/detection/data/images/temp/dilate.jpg",dilate)

    cnts = cv2.findContours(dilate,cv2.RETR_LIST,cv2.CHAIN_APPROX_TC89_KCOS)
    cnts = cnts[0] if len(cnts)==2 else cnts[1]
    ymax=0.0
    wmoy=0.0
    hmoy=0.0
    
    n=len(cnts)
    for cnt in cnts:
        x,y,w,h =cv2.boundingRect(cnt)
        if y>ymax : ymax=y
        wmoy+=w/n
        hmoy+=h/n

    if direction==ReadRirection.col :
        cnts = sorted(cnts,key=lambda x: sortKeyByColumns(x,ymax,hmoy,wmoy),reverse=True)
    if direction==ReadRirection.row:
        cnts = sorted(cnts,key=lambda x: sortKeyByRows(x,ymax,hmoy,wmoy),reverse=True)


    rois=[]
    base_image = image.copy()
    lg=len(cnts)
    for i in range(lg) :
        c=cnts[lg-i-1]
        x,y,w,h = cv2.boundingRect(c)
        if (h>0.5*hmoy and h<max_high) and (w>0.5*wmoy and w<max_wide):
            roi = base_image[y:y+h,x:x+w]
            rois.append(roi)     
            cv2.rectangle(image,(x,y),(x+w,y+h),(0,255,0),2)

    for roi in rois:
        i+=1
        cv2.imwrite(f"./ocr/detection/data/images/temp/rois_{i}.png",roi)
            
    cv2.imwrite(f"./ocr/detection/data/images/temp/bbox_{name}.png",image)
    
    return rois



def smooth(img, type="mean", kernel_size=9):
    if type == "mean":
        return cv2.blur(img, (kernel_size, kernel_size))
    elif type == "median":
        return cv2.medianBlur(img, kernel_size)
    elif type == "gaussian":
        return cv2.GaussianBlur(img, (kernel_size, kernel_size), 0)
    else:
        raise Exception(f"Unsupported smoothing type: {type}")



def get_limits(color=[255,0,0])->Tuple[np.ndarray, np.ndarray]:
    """
    Calcule les limites inférieure et supérieure HSV pour une couleur donnée en BGR.
    """
    c :np.ndarray = np.uint8([[color]])  
    hsvC = cv2.cvtColor(c, cv2.COLOR_BGR2HSV)

    lowerLimit:Tuple[int, int, int] = hsvC[0][0][0] - 15, 50, 50
    upperLimit:Tuple[int, int, int]  = hsvC[0][0][0] + 15, 255, 255

    lowerLimit = np.array(lowerLimit, dtype=np.uint8)
    upperLimit = np.array(upperLimit, dtype=np.uint8)

    return lowerLimit, upperLimit

def save_image(image: MatLike, folder: str, filename: str) -> None:
    """
    Sauvegarde une image dans un dossier spécifié.
    image: Image à sauvegarder au format MatLike.
    folder: Chemin du dossier de sauvegarde.
    filename: Nom du fichier.
    """
    if not os.path.exists(folder):
        os.makedirs(folder)
    output_path: str = os.path.join(folder, filename)
    cv2.imwrite(output_path, image)
    logger.debug(f"Image sauvegardée dans : {output_path}")

def load_and_prepare_image(filename:str)->MatLike:
    """
    filename: chemin du fichier image.
    return: image chargée
    """
    if not os.path.exists(filename):
        raise Exception(f"Le fichier {filename} n'existe pas")
    img = cv2.imread(filename)
    return img

def detect_main_contours(img:MatLike)-> RotatedRect:
    """
    Détecte le contour principal dans une image.
    img : image source.
    return: rectangle correspondant au contour principal.

    """
    gray:MatLike = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    
    blur:MatLike = cv2.GaussianBlur(gray, (5, 5), 0)
    
    canny:MatLike = cv2.Canny(blur, 10, 70)
    
    thresh:MatLike = cv2.adaptiveThreshold(canny, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY_INV, 17, 2)
    
    contours, _ = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cnt = max(contours, key=cv2.contourArea)
    
    rect = cv2.minAreaRect(cnt)
    box = cv2.boxPoints(rect)
    box = np.int0(box)
    return rect

def rotate_image(img:MatLike, rect)->MatLike:
    """
    Effectue la rotation d'une image selon un rectangle détécté.
    img: image source.
    rect: rectangle rotatif détecté.
    return: image après rotation.
    """
    center:Tuple[float,float] = rect[0]
    angle:float = rect[2]
    if rect[1][0] < rect[1][1]:
        angle = 90 - rect[2]
    else:
        angle = rect[2]
    
    M:MatLike = cv2.getRotationMatrix2D(center, angle, 1.0)
    (h, w) = img.shape[:2]
    cos = np.abs(M[0, 0])
    sin = np.abs(M[0, 1])
    new_w = int((w * cos) + (h * sin))
    new_h = int((h * cos) + (w * sin))
    M[0, 2] += (new_w / 2) - center[0]
    M[1, 2] += (new_h / 2) - center[1]
    
    rotated_img:MatLike = cv2.warpAffine(img, M, (new_w, new_h))
    return rotated_img

def detect_and_adjust_blue_zone(img:MatLike)->Tuple[MatLike, int, int, int, int]:
    """
        Détecte et ajuste la zone bleue dans une image.
        img : image source.
        retrun : image ajustée et coordonnées (x, y, w, h) de la zone bleue.
    """
    hsv:MatLike = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    # lower_blue = np.array([100, 50, 50])  
    # upper_blue = np.array([140, 255, 255]) 
    lower_blue, upper_blue=get_limits()
    mask: MatLike= cv2.inRange(hsv, lower_blue, upper_blue)
    
    kernel = np.ones((5, 5), np.uint8)
    mask:MatLike = cv2.erode(mask, kernel, iterations=1)
    mask:MatLike = cv2.dilate(mask, kernel, iterations=1)
    mask:MatLike= cv2.morphologyEx(mask, cv2.MORPH_OPEN, kernel)
    
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cnt :MatLike= max(contours, key=cv2.contourArea)
    x, y, w, h = cv2.boundingRect(cnt)
    
    if y > img.shape[0] / 2:  # Si la zone bleue est en bas
        img = cv2.rotate(img, cv2.ROTATE_180)
        
        # Recalculer les coordonnées après la rotation
        hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        mask = cv2.inRange(hsv, lower_blue, upper_blue)
        mask = cv2.morphologyEx(mask, cv2.MORPH_OPEN, kernel)
        contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        cnt = max(contours, key=cv2.contourArea)
        x, y, w, h = cv2.boundingRect(cnt)

    return img, x, y, w, h



# def check_recto_verso(img:MatLike)->bool:
#     """
#     Vérifie si une image est en recto (avec visage) ou verso
#     img: image cni
#     return: True si recto, sinon False
#     """
#     face_cascade = cv2.CascadeClassifier(cv2.data.haarcascades + 'haarcascade_frontalface_alt.xml')
#     gray:MatLike = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
#     faces = face_cascade.detectMultiScale(gray, scaleFactor=1.1, minNeighbors=5, minSize=(30, 30))
#     recto = len(faces) > 0

#     return recto

def check_recto_verso(img: MatLike) -> bool:
    """
    Vérifie si une image cni est en recto (avec visage) ou verso.
    img: image source
    return: True si recto, sinon False
    """
    mp_face_detection = mp.solutions.face_detection
    with mp_face_detection.FaceDetection(model_selection=1, min_detection_confidence=0.5) as face_detection:
        results = face_detection.process(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
        recto = results.detections is not None
    return recto


def extract_cni(img : MatLike, rect:RotatedRect)->MatLike:
    """
    Extrait la région de la CNI 
    img:Image source
    return: la cni detectée 
    
    """
    box:np.ndarray = cv2.boxPoints(rect)
    box = np.int0(box)
    x, y, w, h = cv2.boundingRect(box)
    # Découper la région d'intérêt 
    cni = img[y:y + h, x:x + w]
    return cni


from typing import List, Optional

from pydantic import  BaseModel, Field


class Prediction(BaseModel) :
    Name:str=Field(default="")
    Confidence:float=Field(default=0.0)

class Predictions(BaseModel) :
  
    EntityId:str=Field(default="")
    Name:str=Field(default="")
    ModelName:str=Field(default="")
    PredictedAt:Optional[str]=Field(default="")
    BestPrediction:str=Field(default="")
    BestPredictionConfidence:float=Field(default=0.0)
    Results:List[Prediction]=[]


import logging
from typing import Dict, List, Tuple
from lxf import settings
from lxf.domain.tables import lxfTable
from lxf.services.measure_time import measure_time
#logger
logger = logging.getLogger('OCR')
fh = logging.FileHandler('./logs/default_ocr.log')
fh.setLevel(settings.get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(settings.get_logging_level())
logger.addHandler(fh)
import regex
import re
from sentence_transformers import SentenceTransformer, util
from sklearn.metrics.pairwise import cosine_similarity
import numpy as np
from lxf.ai.ocr.text_extractor import extract_text_from_image
from pdf2image import convert_from_path
import cv2
import spacy

# model = SentenceTransformer('all-MiniLM-L6-v2')

# with open('/home/lexia/lexia-services/RevolutionAI/Assistants/Conversationnel/chunking_test') as file:
#     essay = file.read()

# single_sentences_list = re.split(r'(?<=[.?!])\s+', essay)
# print(f"{len(single_sentences_list)} sentences were found")
# sentences = [{'sentence': x, 'index': i} for i, x in enumerate(single_sentences_list)]
# print(sentences[:5])

nlp = spacy.load("fr_core_news_lg")

def combine_sentences(sentences, buffer_size=1):
    # Go through each sentence dict
    for i in range(len(sentences)):

        # Create a string that will hold the sentences which are joined
        combined_sentence = ''

        # Add sentences before the current one, based on the buffer size.
        for j in range(i - buffer_size, i):
            # Check if the index j is not negative (to avoid index out of range like on the first one)
            if j >= 0:
                # Add the sentence at index j to the combined_sentence string
                combined_sentence += sentences[j]['sentence'] + ' '

        # Add the current sentence
        combined_sentence += sentences[i]['sentence']

        # Add sentences after the current one, based on the buffer size
        for j in range(i + 1, i + 1 + buffer_size):
            # Check if the index j is within the range of the sentences list
            if j < len(sentences):
                # Add the sentence at index j to the combined_sentence string
                combined_sentence += ' ' + sentences[j]['sentence']

        # Then add the whole thing to your dict
        # Store the combined sentence in the current sentence dict
        sentences[i]['combined_sentence'] = combined_sentence

    return sentences

def sanitize_document(text:str)->str :
    """
    Nettoyage d'un text en lignes , paragraphes..
    """
    # Encadrer les adresses e-mail de guillemets
    regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\s*\.[A-Z|a-z]{2,}\b'
    text = re.sub(regex, lambda match: '"' + match.group(0).replace(' ', '') + '"', text)
    # Encadrer les URL de guillements
    regex=r'\b(?:http|https|www):?[^\s]+[A-Za-z0-9.-]'
    text =re.sub(regex,lambda match:f'"{match.group(0)}"',text)

    #supprime espaces avant ponctuations
    regex=r'\s+([.,!?;:])'
    text = re.sub(regex, r'\1', text)
    # Ajouter un point aux fins de paragraphes
    # regex = r"(?<!\.)\n"
    # text = re.sub(regex, ".\n", text)
    
    #supprime espace apres ponctuations
    regex = r'(?<=[?!])'
    text = re.sub(regex, '\\g<0> ', text, 0, re.MULTILINE)

    # Supprimer les en-têtes/pieds de page
    regex = r"(Page \d+\/\d+\s*)+"
    text = re.sub(regex, "", text)
    # Les caractères blancs consécutifs
    regex = r" {2,}"
    subst=" "
    text = re.sub(regex, subst, text, 0, re.MULTILINE)
    text=re.sub(r'^[ \t]+|[ \t]+$',' ',text)
    #gestion des guillemets
    regex = r'(?<=[^\s])"(?=[^\s])'
    text = re.sub(regex, ' " ', text)
    # Retours à la ligne suivis d'une majuscule

    # regex = r"\n *?(?P<majuscule>[A-ZÀ])"
    # subst=". \\g<majuscule>"
    # text = re.sub(regex, subst, text, 0, re.MULTILINE)

    # regex=r'(?<![.?!])\n *?(?P<majuscule>[A-ZÀ])'
    # subst=". \\g<majuscule>"
    # text = re.sub(regex, ". \\g<majuscule>", text)

    #Les retours à la ligne restants non suivi d'une majuscule
    regex = r"\n *?(?P<minuscule>[a-z0-9àéèêïöôë])"
    subst = r" \g<minuscule>" 
    text = re.sub(regex, subst, text, 0, re.MULTILINE)
        #listes a puces
    # regex=r'\n-'
    # subst='\n•'
    # text=re.sub(regex,subst,text)
    text =re.sub(r'\n{2,}',' ', text)
    

    return text
@measure_time
def segment_text_into_chunks(text, model_name='all-MiniLM-L6-v2', buffer_size=1, breakpoint_percentile_threshold=95)->List[str]:

    model = SentenceTransformer(model_name, tokenizer_kwargs={"clean_up_tokenization_spaces":True})
    text = sanitize_document(text)   
    single_sentences_list = regex.split(r'(?<!\b(?:M\.|Dr\.|Art\.|Prof\.|Mme\.|St\.|Ex\.|etc\.))(?<!\d[.,])(?<=[.?!])\s+', text)

    sentences = [{'sentence': x, 'index': i} for i, x in enumerate(single_sentences_list)]
    sentences = combine_sentences(sentences, buffer_size=buffer_size)
    
    embeddings = model.encode([sentence['combined_sentence'] for sentence in sentences])
    if len(embeddings) > 1:
        distances = [1 - util.cos_sim(embeddings[i], embeddings[i+1])[0].item() for i in range(len(embeddings)-1)]
    else:
        return [text]
    breakpoint_distance_threshold = np.percentile(distances, breakpoint_percentile_threshold)
    indices_above_thresh = [i for i, distance in enumerate(distances) if distance > breakpoint_distance_threshold]

    chunks = []
    start_index = 0
    for index in indices_above_thresh:
        end_index = index + 1  
        chunk = ' '.join([sentences[i]['sentence'] for i in range(start_index, end_index + 1)])
        chunks.append(chunk)
        start_index = end_index + 1

    if start_index < len(sentences):
        chunk = ' '.join([sentences[i]['sentence'] for i in range(start_index, len(sentences))])
        chunks.append(chunk)

    return chunks



def lemmatize_and_extract_entities(text)->Tuple[str, List[Dict[str, str]]]:
    """
    Effectue la lemmatisation tout en détectant les entités nommées.
    """
    doc = nlp(text)
    
    entities = [{"text": ent.text, "label": ent.label_} for ent in doc.ents]
    lemmatized_text = " ".join([token.lemma_ if not token.ent_type_ else token.text for token in doc])
    
    return lemmatized_text, entities



# async def do_default_ocr(pdf_path: str, use_tesseract=True, use_GPU=True) -> list:
#     """
#     Effectue l'OCR sur chaque page d'un PDF et segmente le texte extrait en chunks.
#     """
#     pages = convert_from_path(pdf_path)
#     all_chunks = []

#     for i, page in enumerate(pages):
#         image_path = f"./data/temp/page_{i+1}.jpeg"
#         page.save(image_path, "JPEG")
#         image = cv2.imread(image_path)
#         if image is None:
#             logger.error(f"Impossible de charger l'image de la page {i+1}.")
#             continue

#         try:
#             extracted_text = extract_text_from_image(image, use_tesseract=use_tesseract, use_GPU=use_GPU)
#             if not extracted_text.strip():
#                 logger.warning(f"Aucun texte détecté sur la page {i+1}, tentative avec Tesseract...")
#                 extracted_text = extract_text_from_image(image, use_tesseract=True, use_GPU=False)
#         except Exception as e:
#             logger.error(f"Erreur OCR sur la page {i+1} : {e}")
#             continue

#         # segmentation du texte extrait en chunks
#         chunks = segment_text_into_chunks(extracted_text)
#         all_chunks.extend(chunks)
#         logger.info(f"{len(chunks)} chunks extraits de la page {i+1}.")

#     return all_chunks










import asyncio
import logging
async def try_safe_execute_async(logger:logging.Logger,func,**kargs) :
    try:
        result= await func(**kargs)
        return result
    except Exception as ex:
        #logger.exception(f"Exception occured in {func} with kargs={kargs}: {ex} ")
        logger.exception(f"Exception occured in {func} : {ex} ")
        return None

def try_safe_execute(logger,func,**kargs) :
    try:
        result= func(**kargs)
        return result
    except Exception as ex:
        logger.exception(f"Exception occured in {func} with kargs={kargs}: {ex} ")
        return None
    
def try_safe_execute_asyncio(logger,func,**kargs) :
    try:

        result= asyncio.run(func(**kargs))

        return result
    except Exception as ex:
        logger.exception(f"Exception occured in {func} with kargs={kargs}: {ex} ")
        return None



import logging

from lxf.settings import get_logging_level

logger = logging.getLogger('Text extractor')
fh = logging.FileHandler('./logs/text_extractor.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)

from typing import List, Optional, Tuple

from lxf.services.measure_time import measure_time
import pytesseract
import cv2
from cv2.typing import MatLike
import easyocr


reader = easyocr.Reader(['fr'],gpu=True) 
if reader==None :
    msg ="Obtention impossible d'un easyocr reader"
    print(msg)
    logger.error(msg)
else :
    msg="Easyocr reader pret"
    print(msg)
    logger.info(msg)
cpu_reader = easyocr.Reader(['fr'],gpu=False)
if cpu_reader==None : 
    msg ="Obtention impossible d'un CPU easyocr reader"
    print(msg)
    logger.error(msg)
else :
    msg="CPU Easyocr reader pret"
    print(msg)
    logger.info(msg)    
    
@measure_time
def extract_text_from_image(final: MatLike, use_tesseract: bool = True,use_GPU:bool=True) -> str:
    """
    Extrait le texte d'une image en utilisant EasyOCR ou Tesseract.
    """
            # Telesserac config
        # Add -l LANG[+LANG] to the command line to use multiple languages together for recognition
        #     -l eng+deu
        # quiet to suppress messages
        # pdf This creates a pdf with the image and a separate searchable text layer with the recognized text.
        # Use --oem 1 for LSTM/neural network, --oem 0 for Legacy Tesseract.
        # –psm 3 - Fully automatic page segmentation, but no OSD. (Default)
        # –psm 6 - Assume a single uniform block of text.
        # –psm 11  Use pdftotext for preserving layout for text output
        # Use -c preserve_interword_spaces=1 to preserve spaces
        # hocr to get the HOCR output
        # <?xml version="1.0" encoding="UTF-8"?>
        # <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
        #     "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        # <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
        #  <head>
        #   <title></title>
        #   <meta http-equiv="Content-Type" content="text/html;charset=utf-8"/>
        #   <meta name='ocr-system' content='tesseract 5.0.1-64-g3c22' />
        #   <meta name='ocr-capabilities' content='ocr_page ocr_carea ocr_par ocr_line ocrx_word ocrp_wconf'/>
        #  </head>
        #  <body>
        #   <div class='ocr_page' id='page_1' title='image "images/eurotext.png"; bbox 0 0 640 500; ppageno 0; scan_res 300 300'>
        #    <div class='ocr_carea' id='block_1_1' title="bbox 61 41 574 413">
        #     <p class='ocr_par' id='par_1_1' lang='eng' title="bbox 61 41 574 413">
        #      <span class='ocr_line' id='line_1_1' title="bbox 65 41 515 71; baseline 0.013 -11; x_size 25; x_descenders 5; x_ascenders 6">
        #       <span class='ocrx_word' id='word_1_1' title='bbox 65 41 111 61; x_wconf 96'>The</span>
        #       <span class='ocrx_word' id='word_1_2' title='bbox 128 42 217 66; x_wconf 95'>(quick)</span>
        #       <span class='ocrx_word' id='word_1_3' title='bbox 235 43 330 68; x_wconf 95'>[brown]</span>
        #       <span class='ocrx_word' id='word_1_4' title='bbox 349 44 415 69; x_wconf 94'>{fox}</span>
        #       <span class='ocrx_word' id='word_1_5' title='bbox 429 45 515 71; x_wconf 96'>jumps!</span>
        #      </span>

        # ...

        #      <span class='ocr_line' id='line_1_12' title="bbox 61 385 444 413; baseline 0.013 -9; x_size 24; x_descenders 4; x_ascenders 5">
        #       <span class='ocrx_word' id='word_1_62' title='bbox 61 385 119 405; x_wconf 92'>salta</span>
        #       <span class='ocrx_word' id='word_1_63' title='bbox 135 385 200 406; x_wconf 92'>sobre</span>
        #       <span class='ocrx_word' id='word_1_64' title='bbox 216 392 229 406; x_wconf 83'>o</span>
        #       <span class='ocrx_word' id='word_1_65' title='bbox 244 388 285 407; x_wconf 80'>cdo</span>
        #       <span class='ocrx_word' id='word_1_66' title='bbox 300 388 444 413; x_wconf 92'>preguigoso.</span>
        #      </span>
        #     </p>
        #    </div>
        #   </div>
        #  </body>
        # </html>
        # TSV to get the TSV output :
        # level   page_num        block_num       par_num line_num        word_num        left    top     width   height  conf    text
        # 1       1       0       0       0       0       0       0       640     500     -1
        # 2       1       1       0       0       0       61      41      513     372     -1
        # 3       1       1       1       0       0       61      41      513     372     -1
        # 4       1       1       1       1       0       65      41      450     30      -1
        # 5       1       1       1       1       1       65      41      46      20      96.063751       The
        # 5       1       1       1       1       2       128     42      89      24      95.965691       (quick)
        # 5       1       1       1       1       3       235     43      95      25      95.835831       [brown]
        # 5       1       1       1       1       4       349     44      66      25      94.899742       {fox}
        # 5       1       1       1       1       5       429     45      86      26      96.683357       jumps!
    custom_config = r'--oem 1 --oem 3 --psm 6 '        

    if use_tesseract :
         text=pytesseract.image_to_string(final,config=custom_config)

    else:
        try:
            text = ""
            if use_GPU :
                result = reader.readtext(final, detail=0)
            else : 
                result = cpu_reader.readtext(final,detail=0)
            text = " ".join(result)
        except Exception as ex:
            logger.error(f"Exception pendant l'extraction de texte : {ex}. \nEssai en CPU")
            return extract_text_from_image(final,use_tesseract=False,use_GPU=False)
    return text


import logging
import os
from sentence_transformers import SentenceTransformer, util
import spacy
from transformers import MBartForConditionalGeneration, MBart50TokenizerFast
from torch import cuda

if spacy.prefer_gpu() :
    print("Spacy will use GPU")
else :
    print("Spacy will use CPU")

global nlp_with_vectors
try:
    print("Chargement du modèle SPACY : fr_core_news_lg ")
    nlp_with_vectors = spacy.load("fr_core_news_lg")
except Exception as ex:
    nlp_with_vectors=None
    print("le model SPACY : fr_core_news_lg est obligatoire. Exécuter la ligne de commande suivante: python -m spacy download fr_core_news_lg ")

global nlp
nlp=None
global model_path
model_path='models/nlp/lex-fr.nlp'

global enable_tqdm
enable_tqdm:bool=False

global SET_LOGGING_LEVEL
SET_LOGGING_LEVEL = logging.ERROR

# Chargement du modèle d'embedding SentenceTransformer
global sentence_embedding_model
sentence_embedding_model = SentenceTransformer('all-MiniLM-L6-v2', tokenizer_kwargs={"clean_up_tokenization_spaces": True})

# Chargement du modèle de summarization
device = "cuda" if cuda.is_available() else "cpu"
model_name = "facebook/mbart-large-50"
global text_tokenizer
text_tokenizer = MBart50TokenizerFast.from_pretrained(model_name)
text_tokenizer.src_lang = "fr_XX"
global text_summarization_model
text_summarization_model = MBartForConditionalGeneration.from_pretrained(model_name).to(device)



def get_logging_level():
    """
    Return the current LOGGING LEVEL : DEBUG=10, ERROR=40 
    """
    return SET_LOGGING_LEVEL

def set_logging_level(logging_level)->None:
    """
    Change the current Logging 
    """
    global SET_LOGGING_LEVEL
    SET_LOGGING_LEVEL=logging_level

if os.path.exists("./logs") ==False:
    print("Création du dossier logs")
    os.mkdir("./logs")

def load_model(path:str=None):
    """
    """
    global model_path
    global nlp
    if path!=None and path!=model_path:
        model_path=path        
        if os.path.exists(model_path) :
            print(f"Chargement du modèle depuis: {model_path}")
            nlp=spacy.load(model_path)
        else :
            print(f"Erreur: le modèle est introuvable: {model_path}")
            return None            
    if nlp==None:
        nlp=spacy.load(model_path)
    return nlp


global nlp_title
nlp_title=None
global model_title_path
model_title_path="models/nlp/lex-title-model-trained-V2/model-best"

def load_model_title(path:str=None):
    """
    """
    global model_title_path
    global nlp_title
    if path!=None and path!=model_title_path:
        model_title_path=path        
        if os.path.exists(model_title_path) :
            print(f"Chargement du modèle depuis: {model_title_path}")
            nlp_title=spacy.load(model_title_path)
        else :
            print(f"Erreur: le modèle est introuvable: {model_title_path}")
            return None            
    if nlp_title==None:
        nlp_title=spacy.load(model_title_path)
    return nlp_title



###################################################################

import logging
from lxf.domain.iban import Iban
from lxf.settings import get_logging_level

logger = logging.getLogger('iban extractor')
fh = logging.FileHandler('./logs/iban_extractor.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)
#################################################################


from lxf.ai.ocr.iban_ocr import do_IBAN_Ocr_from_pdf
from lxf.extractors.finance.ibans.iban_analyzer import IbanAnalyzer
from lxf.services.try_safe import try_safe_execute_async


async def extract_data(file_path:str,**kwargs)->tuple[str,list[Iban]]:
    """
    Extrait les données d'un iban 
    """

    if file_path==None: 
        logger.error("Extract Data: file_path ne peut pas etre None")
        return None, None
    text_ocr = await try_safe_execute_async(logger,do_IBAN_Ocr_from_pdf, path_filename=file_path)
    if text_ocr!=None :
        analyzer:IbanAnalyzer= IbanAnalyzer(text_ocr)
        return text_ocr, await try_safe_execute_async(logger,analyzer.do_analyze)
    return text_ocr, None


import json



def correct_prodigy_dataset(input_file, output_file):
    with open(input_file, "r", encoding="utf-8") as f:
        lines = f.readlines()

    corrected_data = []
    for line in lines:
        example = json.loads(line)
        
        # if example.get("answer") == "reject":
        #     continue
        
        if "spans" in example and example["spans"]:
            # filtrer les spans incomplets
            corrected_spans = []
            for span in example["spans"]:
                if all(key in span for key in ["text", "start", "end", "label"]):
                    # supprimer les sauts de ligne et les espaces au début et à la fin du texte du span
                    span["text"] = span["text"].strip()
                    # ajuster les positions start et end 
                    span_start = example["text"].find(span["text"])
                    if span_start != -1: 
                        span["start"] = span_start
                        span["end"] = span_start + len(span["text"])
                        corrected_spans.append(span)
            example["spans"] = corrected_spans
        else:
            continue
        corrected_data.append(example)

    with open(output_file, "w", encoding="utf-8") as f:
        for example in corrected_data:
            f.write(json.dumps(example, ensure_ascii=False) + "\n")

    print(f"Correction terminée. Fichier corrigé enregistré sous : {output_file}")

input_file = "/home/lexia/works/RevolutionAI/Core/Libs-python/src/data/annotations_manual.jsonl"
output_file = "/home/lexia/works/RevolutionAI/Core/Libs-python/src/data/annotations__manual_correct_NoReject.jsonl"
correct_prodigy_dataset(input_file,output_file)


import spacy
from spacy.lang.fr.stop_words import STOP_WORDS
from string import punctuation
import logging
from lxf.settings import get_logging_level, nlp_with_vectors

#logger
logger = logging.getLogger('Text Comparator')
fh = logging.FileHandler('./logs/text_comparator.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)

chiffres ={"0","1","2","3","4","5","6","7","8","9"}
STOP_WORDS.update(chiffres)
symbols={"€","£","|","§","*"}
STOP_WORDS.update(symbols)
punctuations ={punct for punct in punctuation}
STOP_WORDS.update(punctuations)
EMPTY_STR="empty"
JOKER_STR="***"


class TextComparator :
    """
    Compare the internal text to text provide as parameter
                Comparation uses vectorization of token not in stop_word(spacy) and similarity beetwen text to be compared
    """
    def __init__(self,text:str,nlp=None):
        if nlp == None:
            self.nlp_test=nlp_with_vectors
        else : 
            self.nlp_test = nlp        
        self.internal_text=self.sanitize_text_and_clean_stop_words(text)
        # since we cannot compare by cosimilarity an empty string (=null vector), we decide to assign EMPTY_STR as empty string
        if self.internal_text=="" :            
            self.internal_text=EMPTY_STR
        self.internal_doc = self.nlp_test(self.internal_text)
        
    def sanitize_text_and_clean_stop_words(self,text):
        """
        Sanitize the text and retains only token not in STOP_WORDS
        """
        temp_text= self.sanitize_text(text.lower()) 
        retained_text=[word for word in temp_text.split() if word not in STOP_WORDS]
        return  " ".join(retained_text)      
        
    def get_nlp(self) :
        return self.nlp_test
    
    def get_text(self) :
        return self.internal_text 
                                     
    def sanitize_text(self,text:str)->str :
        temp_text=""
        last_char=" "
        # remove extra space
        text=text.replace('–','-').replace('’','\'').replace("\n"," ").strip()
        for char in text:
            if char==" ":
                if not last_char== " ":
                    temp_text+=char
                last_char=char
            else:
                last_char=char
                temp_text+=char
        return temp_text 
    
    def compare_to(self, text:str)->float :
        """
        Compare the internal text to text provide as parameter
            Comparation uses vectorization of token not in stop_word(spacy) and similarity beetwen text to be compared
        return : float value beetwen in -1 and 1 ; 100% means equal 
        """
        text_to_compare = self.sanitize_text_and_clean_stop_words(text)
        if text_to_compare==None :
            return 0.0 
        # since we cannot compare by cosimilarity an empty string (=null vector), we decide to assign EMPTY_STR as empty string
        if text_to_compare=="":
            text_to_compare=EMPTY_STR
        # if internal text is JOKER_STR so return 1 since JOKER replace any char
        elif self.internal_text==JOKER_STR :
            return 1.0
        doc_to_compare = self.nlp_test(text_to_compare)
        if doc_to_compare.has_vector :
            #logger.debug(f"Comparing  {text_to_compare} <= ? => {self.internal_text} ")
            result= self.internal_doc.similarity(doc_to_compare)
            if text_to_compare.startswith(self.internal_text) :
                # si le text à comparer commence par le text interne alors on retourne la valeur max entre 0.8 
                # et le résult du calcul de la cosimilarité
                return max(result,0.8)
            return result
        else :
            return 0.0 

class TextArrayComparator:
    """ 
    Compare array fo text using vectorization and cosimilarity 
    """
    def __init__(self,texts_array,nlp) :
        self.internal_docs=[ TextComparator(text,nlp=nlp) for text in texts_array]        
    
    def compare_to(self, texts_array,cell_limit=-1)->float :
        """
        cell_limit = -1 compare all cells
        otherwise only the first cell_limit cells 
        """
        if not len(texts_array) == len(self.internal_docs) :
            #logger.debug(f"empty texts array found")
            return 0.0
        sum=0.0
        if cell_limit==-1 :
            for i,cptor in enumerate(self.internal_docs) :
                sum+=cptor.compare_to(texts_array[i])
            lg=float(len(self.internal_docs))
        else:
            for i,cptor in enumerate(self.internal_docs[:cell_limit]) :
                sum+=cptor.compare_to(texts_array[i])
            lg=float(cell_limit)
        #logger.debug(f"similarity={sum}/{lg}={sum/lg}")
        return sum/lg


from pydantic import BaseModel,Field

class KeyWord(BaseModel):
	word:str=Field(default="")
	freq:float=Field(default=0.0)
	class Config:        
		json_schema_extra = {
			"example": {
				"word":'facturation',
				"freq":0.35
			}
		}



import logging
from typing import Dict, List, Tuple
from lxf import settings

from lxf.services.measure_time import measure_time
#logger
logger = logging.getLogger('Text Analysis')
fh = logging.FileHandler('./logs/default_text_analysis.log')
fh.setLevel(settings.get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(settings.get_logging_level())
logger.addHandler(fh)
import regex
import re
from sentence_transformers import util

import numpy as np


from torch import cuda
from lxf.settings import load_model_title, nlp_with_vectors , text_summarization_model , sentence_embedding_model , text_tokenizer


def combine_sentences(sentences, buffer_size=1):
    # Go through each sentence dict
    for i in range(len(sentences)):

        # Create a string that will hold the sentences which are joined
        combined_sentence = ''

        # Add sentences before the current one, based on the buffer size.
        for j in range(i - buffer_size, i):
            # Check if the index j is not negative (to avoid index out of range like on the first one)
            if j >= 0:
                # Add the sentence at index j to the combined_sentence string
                combined_sentence += sentences[j]['sentence'] + ' '

        # Add the current sentence
        combined_sentence += sentences[i]['sentence']

        # Add sentences after the current one, based on the buffer size
        for j in range(i + 1, i + 1 + buffer_size):
            # Check if the index j is within the range of the sentences list
            if j < len(sentences):
                # Add the sentence at index j to the combined_sentence string
                combined_sentence += ' ' + sentences[j]['sentence']

        # Then add the whole thing to your dict
        # Store the combined sentence in the current sentence dict
        sentences[i]['combined_sentence'] = combined_sentence

    return sentences

def sanitize_document(text:str)->str :
    """
    Nettoyage d'un text en lignes , paragraphes..
    """
    # Supprimer les espaces entre les lettres d'un mot
    text = re.sub(r"([A-Za-z]) ([A-Za-z]) ", r"\g<1>\g<2>", text)
    # Encadrer les adresses e-mail de guillemets
    regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\s*\.[A-Z|a-z]{2,}\b'
    text = re.sub(regex, lambda match: '"' + match.group(0).replace(' ', '') + '"', text)
    # Encadrer les URL de guillements
    regex=r'\b(?:http|https|www):?[^\s]+[A-Za-z0-9.-]'
    text =re.sub(regex,lambda match:f'"{match.group(0)}"',text)

    #supprime espaces avant ponctuations
    regex=r'\s+([.,!?;:])'
    text = re.sub(regex, r'\1', text)
    # Ajouter un point aux fins de paragraphes
    # regex = r"(?<!\.)\n"
    # text = re.sub(regex, ".\n", text)
    
    #supprime espace apres ponctuations
    regex = r'(?<=[?!])'
    text = re.sub(regex, '\\g<0> ', text, 0, re.MULTILINE)

    # Supprimer les en-têtes/pieds de page
    regex = r"(Page \d+\/\d+\s*)+"
    # text = re.sub(regex, "", text)
    # Les caractères blancs consécutifs
    regex = r" {2,}"
    subst=" "
    text = re.sub(regex, subst, text, 0, re.MULTILINE)
    # text=re.sub(r'^[ \t]+|[ \t]+$',' ',text)
    #gestion des guillemets
    regex = r'(?<=[^\s])"(?=[^\s])'
    text = re.sub(regex, ' " ', text)
    # Retours à la ligne suivis d'une majuscule

    # regex = r"\n *?(?P<majuscule>[A-ZÀ])"
    # subst=". \\g<majuscule>"
    # text = re.sub(regex, subst, text, 0, re.MULTILINE)

    # regex=r'(?<![.?!])\n *?(?P<majuscule>[A-ZÀ])'
    # subst=". \\g<majuscule>"
    # text = re.sub(regex, ". \\g<majuscule>", text)

    #Les retours à la ligne restants non suivi d'une majuscule
    # regex = r"\n *?(?P<minuscule>[a-z0-9àéèêïöôë])"
    # subst = r" \g<minuscule>" 
    # text = re.sub(regex, subst, text, 0, re.MULTILINE)
        #listes a puces
    # regex=r'\n-'
    # subst='\n•'
    # text=re.sub(regex,subst,text)
    text = re.sub(r'\n{3,}', '\n\n', text)
    return text

def is_validate_title(text: str, context: str = "") -> tuple[bool, str]:
    """
    Verifie qu'il s'agit bien d'un titre
    """
    #Verifier qu'il y a au moins une majuscule
    rule_has_one_capital_at_least=r"[A-Z]" 
    matches = re.search(rule_has_one_capital_at_least,text)
    if matches==None :
        return False, None
    
    # Vérification de la longueur du contexte
    rule0=r"\n*.{1,60}\n"
    # Vérifier si le titre commence par une suite de chiffres
    rule1 = r"[0-9]{4,}[,.\s]?[A-Za-z]*"
    # Détecter une suite de caractères spéciaux
    rule2 = r"^[\.\-\%\&\+\±\/]{2,}" # \(\)  on enlève ce cas car on a des titres par exemple : ASSURANCE(S) 
    # Détecter un titre commençant par exactement deux lettres sauf exceptions
    rule3 = r"^\b([a-zA-Z]{2})\b"
    # Détecter si un titre commence par une ponctuation
    rule_punctuation = r"^[^\w\d]"
    # Vérifier si un mot contient trop de consonnes
    rule_consonnes = r"[BCDFGHJKLMNPQRSTVWXZbcdfghjklmnpqrstvwxz$\[\)\(\!\?\+\&\%\#\;\:\@\,\'\"\\x]{4,}"

    exceptions = {"le", "la", "un", "une", "du", "de", "au", "en"}
    
    if len(text.strip()) < 3:
        return False, None
    # Vérification du contexte
    if re.match(rule0, context) is None:
        return False, None

    mots = text.split()
    if not mots:  
        return False, None

    premier_mot = mots[0] 

    # Rejeter le titre si le premier mot est invalide
    if re.search(rule1, premier_mot):  
        return False, None
    if re.search(rule2, premier_mot): 
        return False, None
    if (re.search(rule_consonnes,mots[0]) != None):
        return False ,None
    if re.match(rule_punctuation, premier_mot):  
        return False, None

    match_rule3 = re.match(rule3, premier_mot, re.IGNORECASE | re.MULTILINE)
    if match_rule3:
        first_word = match_rule3.group(1).lower()
        if first_word not in exceptions:
            return False, None

    # Si le premier mot est valide, on garde tout le titre
    return True, text

@measure_time
def extract_titles(text: str) -> List[str]:
    """
    """
    text=sanitize_document(text)
    nlp_title = load_model_title()
    titles = []
    if nlp_title !=None :
        doc=nlp_title(text)
        for ent in doc.ents :
            if ent.label_=="TITRE" :
                context:str=text[ent.start_char:ent.start_char+80]
                valide , valide_tile = is_validate_title(ent.text,context)
                if valide :
                    titles.append({ 
                        "titre":valide_tile,
                        "debut":ent.start_char,
                        "fin":ent.end_char
                    })
    else :
        title_regex = r"^(?!\s*(P\s*[Aa]\s*[Gg]\s*[Ee])\b)(?![A-Z0-9./\- ]{2,10}$)[0-9.]{0,}(I|II|III|IV|V|VI)*[)\.\- ]{0,} *[A-Z][A-Z’']*[a-z’']*[^!?,:;.]{1,60}$"
        for match in re.finditer(title_regex, text, re.MULTILINE):
            titles.append({
                "titre": match.group(0).strip(),
                "debut": match.start(),
                "fin": match.end()
            })
    nlp_title=None
    return titles


def decouper_text_par_titres(text: str) -> List[Dict[str, str]]:
    titres_positions = extract_titles(text)
    text = sanitize_document(text)
    sections = []
    if titres_positions and titres_positions[0]["debut"] > 0:
        sections.append({
            "titre": "Introduction (avant le premier titre)",
            "contenu": text[:titres_positions[0]["debut"]].strip()
        })
    titres_fusionnes = titres_positions[:]
    for i, titre_position in enumerate(titres_fusionnes):
        debut_contenu = titre_position["fin"]
        if i < len(titres_fusionnes) - 1:
            fin_contenu = titres_fusionnes[i + 1]["debut"]
        else:
            fin_contenu = len(text)
        contenu = text[debut_contenu:fin_contenu].strip()
        if not contenu:
            if i < len(titres_fusionnes) - 1:
                titres_fusionnes[i + 1]["titre"] = titre_position["titre"] + "\n" + titres_fusionnes[i + 1]["titre"]
            continue  
        sections.append({
            "titre": titre_position["titre"],
            "contenu": contenu
        })
    return sections


@measure_time
def segment_text_into_chunks(text, buffer_size=32, breakpoint_percentile_threshold=90)->List[str]:
    #model = sentence_embedding_model
    text = sanitize_document(text)   
    paragraphs = regex.split(r'\n{2,}', text)

    all_chunks = []

    for paragraph in paragraphs:
        paragraph = paragraph.strip()
        if not paragraph:
            continue
        single_sentences_list = regex.split(
            r'(?<!\b(?:M\.|Dr\.|Art\.|Prof\.|Mme\.|St\.|Ex\.|etc\.))(?<!\d\.)(?<=[.?!])\s+', 
            paragraph
        )
        sentences = [{'sentence': x, 'index': i} for i, x in enumerate(single_sentences_list)]
        sentences = combine_sentences(sentences, buffer_size=buffer_size)
        embeddings = sentence_embedding_model.encode([sentence['combined_sentence'] for sentence in sentences])
        if len(embeddings) > 1:
            distances = [1 - util.cos_sim(embeddings[i], embeddings[i + 1])[0].item() for i in range(len(embeddings) - 1)]
        else:
            all_chunks.append(paragraph)
            continue
        breakpoint_distance_threshold = np.percentile(distances, breakpoint_percentile_threshold)
        indices_above_thresh = [i for i, distance in enumerate(distances) if distance > breakpoint_distance_threshold]
        start_index = 0
        for index in indices_above_thresh:
            end_index = index + 1
            chunk = ' '.join([sentences[i]['sentence'] for i in range(start_index, end_index + 1)])
            all_chunks.append(chunk)
            start_index = end_index + 1
        if start_index < len(sentences):
            chunk = ' '.join([sentences[i]['sentence'] for i in range(start_index, len(sentences))])
            all_chunks.append(chunk)

    return all_chunks

def lemmatize_and_extract_entities(text)->Tuple[str, List[Dict[str, str]]]:
    """
    Effectue la lemmatisation tout en détectant les entités nommées.
    """
    nlp = nlp_with_vectors
    doc = nlp(text)
    
    entities = [{"text": ent.text, "label": ent.label_} for ent in doc.ents]

    words_to_preserve={"madame","monsieur","mademoiselle","Madame","Monsieur","Mademoiselle","Mesdames", "Messieurs","Docteur", "Professeur","RIB", "IBAN","km", "kg", "€", "$"}

    lemmatized_text = " ".join([
        token.text if token.text in words_to_preserve or token.ent_type_ else token.lemma_
        for token in doc
    ])    
    nlp=None
    return lemmatized_text, entities


def split_large_chunk(chunk: str, max_length: int, tokenizer) -> List[str]:
    """
    """
    if not chunk.strip():
        return []
    tokens = tokenizer.encode(chunk, truncation=False, add_special_tokens=False)
    if not tokens:
        return []
    chunks = []
    start = 0
    while start < len(tokens):
        end = min(start + max_length, len(tokens))

        if end < len(tokens):
            for i in range(end - 1, start - 1, -1):  
                if tokenizer.decode([tokens[i]]) in {".", "!", "?"}:
                    end = i + 1  
                    break

        subchunk_tokens = tokens[start:end]
        subchunk = tokenizer.decode(subchunk_tokens, skip_special_tokens=True)
        chunks.append(subchunk.strip())
        start = end
    return chunks

def generate_summary(text: str, max_length=300, min_length=150, num_beams=4) -> str:
    """
    """
    if len(text.split()) < 30:  
        return text 
    device = "cuda" if cuda.is_available() else "cpu"
    inputs = text_tokenizer(text, return_tensors="pt", truncation=True, max_length=max_length).to(device)

    summary_ids = text_summarization_model.generate(
        inputs["input_ids"],
        num_beams=num_beams,
        max_length=max_length,
        min_length=min_length,
        early_stopping=True,
        no_repeat_ngram_size=2,
        forced_bos_token_id=text_tokenizer.lang_code_to_id["fr_XX"] 
    )
    return text_tokenizer.decode(summary_ids[0], skip_special_tokens=True)

def summarize_chunks(text: str, summary_max_length: int = 1024) -> List[str]:
    """
    """
    logger.debug("Segmentation du texte en chunks...")
    chunks = segment_text_into_chunks(text)

    logger.debug("Division des chunks trop grands...")
    all_subchunks = []
    for chunk in chunks:
        subchunks = split_large_chunk(chunk, max_length=summary_max_length, tokenizer=text_tokenizer)
        all_subchunks.extend(subchunks)

    logger.debug("Generation des resumes intermediaires...")
    chunk_summaries = []
    for i, subchunk in enumerate(all_subchunks):
        try:
            summary = generate_summary(subchunk)
            chunk_summaries.append(f"- {summary}")  
            logger.debug(f"Resume intermediaire {i + 1}/{len(all_subchunks)} genere.")
        except Exception as e:
            logger.error(f"Erreur lors du resume intermediaire pour le chunk {i + 1}: {e}")
            # chunk_summaries.append(f"Erreur lors du résumé.")
    return chunk_summaries


from typing import List
from xml.dom.minidom import Entity
from pydantic import BaseModel,Field

from lxf.domain.tables import lxfTable
from enum import Enum

class LoanTemplate(Enum):
    BPOP=1
    CAISSE_EPARGNE=2


class Montant(BaseModel):
    """
    Montant et devise
    """
    valeur:float=Field(default=0.0)
    devise:str=Field(default="EUR")
    class Config :
        json_schema_extra ={
            "exemple":{
                "valeur":123456.57,
                "devise":"EUR"
            }
        }

class MontantDivers(BaseModel):
    """
    Frais 
    """
    txt_montant:str=""
    nature:str=Field(default="Montant divers)")
    montant:Montant=Field(default=Montant())
    class Config:
        json_schema_extra={
            "exemple":{
                "label":"Garantie (PPD)",
                "nature":"garantie (PPD)",
                "montant":{"valeur":3808.51,"devise":"EUR"}
            }
        }

class Personne(BaseModel):
    """
    Personne Morale ou Persone Physique
    """
    est_personne:bool=Field(default=False)
    civilite:str =Field("")
    prenoms:str=Field("")
    raison_sociale:str=Field("")
    statut_organisation:str=Field("")
    siret:str=Field("")
    siren:str=Field("")
    nom:str=Field("")
    date_naissance:str=Field("")
    adresse:str=Field("")
    code_postal:str=Field("")
    ville:str=Field("")
    metier:str=Field("")
    situation_maritale:str=Field("")
    regime_matrimoniale:str=Field("")
    class Config :
        json_schema_extra ={
            "exemple" :{
                "est_personne":True,
                "cilvilite":"M",
                "prémon":"Jean Marc Antoine",
                "nom":"LEMARCHAND",
                "metier":"cadre fonction publique",
                "situation_maritale":"marié",
                "regime_maritale":"Séparation de biens",
                "adresse":"140, boulevard des Capucines",
                "code_postal":"01234",
                "ville":"Bon Vivre sur Mer et Montagne"
            }
        }

class Preteur(Personne):    
    """
    Organisme prêteur dans le cadre d'un contrat de prêt
    """
    represente_par:List[Personne]=[]
    class Config :
        json_schema_extra ={
            "exemple" :{
                "est_personne":False,
                "raison_sociale":"Banque Crédit Sécure",
                "statut_organisation":"Société Coopérative",
                "siren":"111 222 333",
                "adresse":"140, boulevard des Capucines",
                "code_postal":"01234",
                "ville":"Bon Vivre sur Mer et Montagne",
                "represente_par":[]
            }
        }

class Emprunteur(Personne):
    """
    Personne physique ou morale bénéficière d'un prêt bancaire
    """
    represente_par:List[Personne]=[]
    class Config :
        json_schema_extra ={
            "exemple" :{
                "est_personne":True,
                "civilite":"Mme",
                "premons":"Marie Françoise",
                "nom":"FRIDOMME",
                "metier":"commerçante",
                "situation_maritale":"divorcée",
                "regime_maritale":"",
                "adresse":"140, boulevard des Capucines",
                "code_postal":"01234",
                "ville":"Bon Vivre sur Mer et Montagne",
                "represente_par":[]
            }
        }

class ElementPlanFinancement(BaseModel):
    """
    Element d'un Plan de Financement d'un Prêt
    """
    nature:str=Field("")
    montant:Montant=Field(default=Montant())
    class Config:
        json_schema_extra ={
            "exemple":{
                "nature":"apport",
                "montant":{
                    "valeur":125000.00,
                    "devise":"EUR"
                    }
            }
        }
 
class ProgrammeFinancier(BaseModel):
    """
    Programme Financier d'un prêt
    plan_financement : List of ElementPlanFinancement
    montant_programme : Montant du programme (valeur,devise)
    """
    plan_financement:List[ElementPlanFinancement]=[]
    montant_programme:Montant=Field(default=Montant())
    class Config:
        json_schema_extra={
            "exemple":{
                "plan_financment":[
                    {"nature":"apport",
                     "montant":{
                         "valeur":125000.00,
                         "devise":"EUR"
                     }
                    },
                    {"nature":"subvention",
                     "montant":{
                         "valeur":0.00,
                         "devise":"EUR"
                     }
                    } ,
                    {"nature":"Prêt(s) BPBFC sollicité(s)",
                     "montant":{
                         "valeur":730000.00,
                         "devise":"EUR"
                     }
                    }                                       
                ],
                "montant_programe":{
                    "valeur":855000.00,
                    "devise":"EUR"
                }
            }
        }

class TauxAnnuelEffectifGlobal(BaseModel):
    """
    Taux annuel Effectif Global TAEG
    """
    titre:str=Field("")
    sigle:str=Field("")
    description:str=Field(" ")
    taux_pourcentage:float=Field(0.0)
    taux_periode:float=Field(0.0)
    base_periode:str=Field(default="mensuelle")
    class Config:
        json_schema_extra={
            "exemple":{
                "taux_pourcentage":2.00,
                "base_periode":"mensuelle"
            }
        }
class Domiciliation(BaseModel):
    """
    Domiciliation du compte à créditer
    """
    debiteur:str=Field("")
    compte:str=Field("")
    creancier:str=Field("")
    class Config:
        json_schema_extra={
            "exemple":{
                "crediteur":"Emprunteur",
                "compte":"compte n° 123456789",
                "beneficiaire":"BANQUE CREDIT"
            }
        }

class Frais(BaseModel):
    """
    Frais 
    """
    nature:str=Field(default="Frais Prise de garantie (Hypohteque)")
    montant:Montant=Field(default=Montant())
    class Config:
        json_schema_extra={
            "exemple":{
                "nature":"Frais prise de garantie (PPD)",
                "montant":{"valeur":3808.51,"devise":"EUR"}
            }
        }

class MontantTotalDuEmprunteur(BaseModel):
    """
    Total montant dû par l'emprunteur
    montant_total_credit:Montant total du crédit (valeur,devise)
    cout_total_credit: Cout total du Credit (valeur,devise)
    montant_total_interets: Montant Total des intérets (valeur, devise)
    montant_total_assurance_obligatoire:Montant total des assurances obligatoires (valeur,devise)
    frais: liste de Frais
    montant_total_du_emprunteur: Montant total dû par l'emprunteur (valeur,devise)
    """
    montant_total_credit:Montant=Field(default=Montant())
    cout_total_credit:Montant=Field(default=Montant())
    montant_total_interets:Montant=Field(default=Montant())
    montant_total_assurance_obligatoire:Montant=Field(default=Montant())
    frais:List[Frais]=[]
    montants_divers:List[MontantDivers]=[]
    montant_total_du_emprunteur:Montant=Field(default=Montant())
    class Config:
        json_schema_extra={
            "exemple":{
                "montant_total_credit":{"valeur":730000.00,"devise":"EUR"},
                "cout_total_credit":{"valeur":197342.08,"devise":"EUR"},
                "montant_total_interets":{"valeur":147327.00,"devise":"EUR"},
                "montant_total_assurance_obligatoire":{"valeur":45150.54,"devise":"EUR"},
                "frais":[
                        {
                            "nature":"Frais prise de garantie (PPD)",
                            "montant":{"valeur":3808.51,"devise":"EUR"}
                        },
                        {
                            "nature":"Frais prise de garantie (Hypotheque)",
                            "montant":{"valeur":556.03,"devise":"EUR"}
                        },
                        {
                            "nature":"Frais de dossier",
                            "montant":{"valeur":500.00,"devise":"EUR"}
                        }
                ],
                "montant_total_du_emprunteur":{"valeur":927342.08,"devise":"EUR"}
            }
        }

class PeriodeAmortissement(BaseModel):
    """
    Période d'amortissement d'un prêt
    type_echeance: Type échéance constante
    duree_echeances:la durée des échéances(mensuelle)
    type_taux_debiteur:type de taux débiteur (Fixe)
    taux_debiteur_pourcentage:1.5 
    montant_echeance_sans_assurance:(valeur,devise)
    """
    numero:int=Field(0)
    type_echeance:str=Field(default="constante")
    duree_echeances:str=Field(0)
    periodicite_echeance:str=Field(default="mensuelle")
    type_taux_debiteur:str=Field(default="fixe")
    taux_debiteur_pourcentage:float=Field(0.0)
    montant_echeances:List[MontantDivers] = []
    assurance:Montant=Montant()
    accessoire:Montant=Montant()
    class Config:
        scheam_extra={
            "exemple":{
                "numero":1,
                "type_echeance":"constante",
                "duree_echeances":"8",
                "periodicite_echeance":"mensuelle",
                "type_taux_debiteur":"fixe",
                "taux_debiteur_pourcentage":1.5,
                "montant_echeances":[{"nature":"Montant de l'échéance sans assurance","montant":{"valeur":2000,"devise":"EUR"}}]
            }
        }

class CaracteristiquePret(BaseModel):
    """
    Caractéristique d'un prêt bancaire
    """
    nature:str=Field("")
    numero:str=Field("")
    montant:Montant=Field(default=Montant())
    duree_mois:int = Field(0)
    class Config:
        json_schema_extra={
            "exemple":{
                       "nature":"Pret tout Habitat",
                       "numero":"08801143" ,
                       "montant":{"valeur":730000.00,"devise":"EUR"},
                       "duree_mois":300
                    }
        }
class CaracteristiquesPret(BaseModel):
    """
    Toutes les caractérisitiques d'un pret bancaire
    caracterisitiques : Liste de CrarcterisitiquePret
    """
    #caracteristiques:List[CaracteristiquePret]=[]    
    description:str=Field("")
    numero_pret:str=Field("")
    is_modulable:bool = Field(False)
    montant:str=Field("")
    taux:TauxAnnuelEffectifGlobal=TauxAnnuelEffectifGlobal()
    domiciliation:Domiciliation=Domiciliation()
    montant_total_du_emprunteurs:MontantTotalDuEmprunteur=MontantTotalDuEmprunteur()
    periodes_amortissement:List[PeriodeAmortissement]=[]
    class Config:
        json_schema_extra={
            "exemple":{
                "description":"Prêt Tout Habitat (N°08801143) : 730 000.00 EUR sur 300 mois",
                "numero_pret":"08801143",
                "is_modulable":True,
                "montant":"730 000.00 EUR",
                "domicialiation":{
                    "crediteur":"Emprunteur",
                    "compte":"compte n° 123456789",
                    "beneficiaire":"BANQUE CREDIT"
                }
            }
        }

class Amortissement(BaseModel):
    """
    Amortissement du crédit
    periodes: Liste de PeriodeAmortissement
    """
    periodes:List[PeriodeAmortissement]=[]
    class Config:
        json_schema_extra={
            "exemple":
            {
                "periodes":[
                    {   "type_echeance":"constante",
                        "duree_echeances":8,
                        "periodicite_echeance":"mensuelle",
                        "type_taux_debiteur":"fixe",
                        "taux_debiteur_pourcentage":1.5,
                        "montant_echeance_sans_assurance":{"valeur":2000,"devise":"EUR"}
                    }
                ]
            }
        }

class Assurance(BaseModel):
    """
    Assurance 
    """
    description:str=Field("")
    reference:str=Field("")
    delegation:str=Field("")
    souscripteur:str=Field("")
    nom_naissance_souscripteur:str=Field("")
    date_naissance_souscripteur:str=Field("")
    ville_naissance_souscripteur:str=Field("")
    garanties:str=Field("")
    type_assureur:str=Field("")
    quotite_pret_assure:str=Field("")
    options_choisies:str=Field("")
    hauteur_pourcentage:float=Field(0.0)
    assureur:str=Field("")
    assureur_group:str=Field("")
    couverture:str=Field("")
    tableau_prets:lxfTable=lxfTable()
    class Config:
        arbitrary_types_allowed = True
        json_schema_extra={
            "exemple":{
                "description":"Assurance Déléguée Décés -PTIA I.T.",
                "delegation":"delegation text",
                "souscripteur":"M Alain DUPOND né le 01/01/2001 à MONCHEZMOI",
                "hauteur_pourcentage":100.00,
                "assureur":"ASSURANCE GENERALE",
                "couverture":"Prêt Tout Habitat (N°01234567) : 730 000.00 EUR sur 300 mois"
            },
            "exemple 2":{
                "description":"Cnp Assurance - Bpce Vie",
                "souscripteur":"M Alain DUPOND",
                "date_naissance_souscripteur":"12/05/1996",
                "assureur":"ASSURANCE GENERALE",
                "ville_naissance_souscripteur":"Belleville"
            }
        }

class Warranty(BaseModel):
    """
    Garantie d'un prêt (Hypothèque, PPD, Caution)
    """
    label:str=Field(default="")
    rang:str=Field(default="0")
    hauteur:str=Field(default="0.0")
    duree:str=Field(default="")
    duree_limitee:str=Field(default="")
    duree_unite:str=Field(default="")
    duree_limitee_unite:str=Field(default="")
    text:str=Field(default="")
    cout_approximatif:str=Field(default="")
    office_notarial:str=Field(default="")
    adresse_office_notarial:str=Field(default="")
    code_postal_office_notarial:str=Field(default="")
    ville_office_notarial:str=Field(default="")
    numero_pret:str=Field(default="")
    ref_cadastrales:str= Field(default="")
    adresse_immeuble_cadastre:str=Field("")
    code_poste_immeuble_cadastre:str=Field("")
    ville_immeuble_cadatre:str=Field("")
    text_adresse_immeuble_cadastre:str=Field("")
    class Config:
        json_schema_extra={
            "exemple":{
                "label":"Hypothèque",
                "rang":1,
                "hauteur":"130 000,00 EUR",
                "cout_approximatif":"1 365,00 EUR",
                "text":"Hypothèque à hauteur de 130 000.00 EUR de rang 1 .... ",
                "numero_pret":"08909399",
                "ref_cadastrales":"SECTION AB NUMEROS 123 ; 456 ; 78 et LOT 81"
                
            }
        }

class ObjetFinancement(BaseModel):
    """
    Objet du financement
    """
    titre:str=""
    nature:str=""
    usage:str=""
    description:str=""
    adresse:str=""
    code_postal:str=""
    ville:str=""

class ClausesParticulieres(BaseModel):
    """
    Clause(s) Particulière(s) rattachée(s) à un prêt
    """
    titre:str=Field("Aucune")
    numero_pret:str=Field("")
    text:str=Field("")
    class Config:
        json_schema_extra={
            "exemple":{
                "text":"L'Emprunteur peut, à tout moment, rembourser en totalité ou en partie le montant du présent prêt.....",
                "numero_pret":"08909399"
            }
        }

class Pret(BaseModel):
    """
    Pret
    """
    num_suivi:str=Field("")
    header_offre:str=Field("")
    header_ref:str=Field("")
    header_emprunteur:str=Field("")
    header_responsable_suivi:str=Field("")
    header_num_etude:str=Field("")
    date_emission_offre:str=Field("")
    objets_financement:List[ObjetFinancement]=[]
    preteur: Preteur = Preteur()
    emprunteurs:List[Emprunteur]=[]
    cautions:List[Emprunteur]=[]
    programme_financier:ProgrammeFinancier=ProgrammeFinancier()
    list_prets:List[CaracteristiquePret]=[]
    caracteristiques_prets:List[CaracteristiquesPret]=[]
    #domiciliation:Domiciliation=Domiciliation()
    amortissement_credit:Amortissement=Amortissement()
    # montant_total_du_emprunteurs:MontantTotalDuEmprunteur=MontantTotalDuEmprunteur()
    delegations_assurances:List[Assurance]=[]
    assurances:List[Assurance]=[]
    garanties:List[Warranty]=[]
    clauses_particulieres:List[ClausesParticulieres]=[]
    tableau_amortissement:lxfTable=lxfTable()
    class Config:        
        arbitrary_types_allowed = True
        json_schema_extra = {
            "example": {
                "description":"Achat immobilier et travaux maison individuelle ....",
                "usage": "Résidence principale emprunteur",
                "preteur":{
                    "est_personne":False,
                    "raison_sociale":"Banque Crédit Sécure",
                    "statut_organisation":"Société Coopérative",
                    "siren":"111 222 333",
                    "adresse":"140, boulevard des Capucines",
                    "code_postal":"01234",
                    "ville":"Bon Vivre sur Mer et Montagne"
                    },
                "emprunteurs":[
                    {
                    "est_personne":True,
                    "premons":"Marie Françoise",
                    "nom":"FRIDOMME",
                    "metier":"commerçante",
                    "situation_maritale":"divorcée",
                    "regime_maritale":"",
                    "adresse":"140, boulevard des Capucines",
                    "code_postal":"01234",
                    "ville":"Bon Vivre sur Mer et Montagne"
                    },
                    {
                    "est_personne":True,
                    "cilvilite":"M",
                    "prémon":"Jean Marc Antoine",
                    "nom":"LEMARCHAND",
                    "metier":"cadre fonction publique",
                    "situation_maritale":"marié",
                    "regime_maritale":"Séparation de biens",
                    "adresse":"140, boulevard des Capucines",
                    "code_postal":"01234",
                    "ville":"Bon Vivre sur Mer et Montagne"
                    }
                ],
                "programme_financier" : {
                                            "plan_financment":[
                                                {"nature":"apport",
                                                "montant":{
                                                    "valeur":125000.00,
                                                    "devise":"EUR"
                                                }
                                                },
                                                {"nature":"subvention",
                                                "montant":{
                                                    "valeur":0.00,
                                                    "devise":"EUR"
                                                }
                                                } ,
                                                {"nature":"Prêt(s) BPBFC sollicité(s)",
                                                "montant":{
                                                    "valeur":730000.00,
                                                    "devise":"EUR"
                                                }
                                                }                                       
                                            ],
                                            "montant_programe":{
                                                "valeur":855000.00,
                                                "devise":"EUR"
                                            }
                                        },
                "caracteristes_prets":[
                                    {
                                        "description":"Prêt Tout Habitat (N°08801143) : 730 000.00 sur 300 mois",
                                        "caracteristiques":[
                                            {
                                            "nature":"Pret tout Habitat",
                                            "numero":"08801143" ,
                                            "montant":{"valeur":730000.00,"devise":"EUR"},
                                            "duree_mois":300
                                            } 
                                        ]                                        
                                    }],
                "amortissement_credit":{
                                        "periodes":
                                        [
                                            {   "type_echeance":"constante",
                                                "duree_echeances":8,
                                                "periodicite_echeance":"mensuelle",
                                                "type_taux_debiteur":"fixe",
                                                "taux_debiteur_pourcentage":1.5,
                                                "montant_echeance_sans_assurance":{"valeur":2000,"devise":"EUR"}
                                            }
                                        ]
                                    },
                "montant_total_du_emprunteurs":{
                                                "montant_total_credit":{"valeur":730000.00,"devise":"EUR"},
                                                "cout_total_credit":{"valeur":197342.08,"devise":"EUR"},
                                                "montant_total_interets":{"valeur":147327.00,"devise":"EUR"},
                                                "montant_total_assurance_obligatoire":{"valeur":45150.54,"devise":"EUR"},
                                                "frais":[
                                                        {
                                                            "nature":"Frais prise de garantie (PPD)",
                                                            "montant":{"valeur":3808.51,"devise":"EUR"}
                                                        },
                                                        {
                                                            "nature":"Frais prise de garantie (Hypotheque)",
                                                            "montant":{"valeur":556.03,"devise":"EUR"}
                                                        },
                                                        {
                                                            "nature":"Frais de dossier",
                                                            "montant":{"valeur":500.00,"devise":"EUR"}
                                                        }
                                                ],
                                                "montant_total_du_emprunteur":{"valeur":927342.08,"devise":"EUR"}
                                            },
                "taeg":{
                        "taux_pourcentage":2.00,
                        "base_periode":"mensuelle"
                        },
                "delegations_assurances":[
                    {
                            "assurance":{
                            "description":"Assurance Déléguée Décés -PTIA I.T.",
                            "souscripteur":"M Alain DUPOND né le 01/01/2001 à MONCHEZMOI",
                            "hauteur_pourcentage":100.00,
                            "assureur":"ASSURANCE GENERALE",
                            "couverture":"Prêt Tout Habitat (N°01234567) : 730 000.00 EUR sur 300 mois"
                            }
                    }
                ],
                "domiciliation":{
                            "debiteur":"Emprunteur",
                            "compte":"compte n° 123456789",
                            "creancier":"BANQUE CREDIT"
                          }               
            }
        }



import logging
import os
from lxf.settings import get_logging_level

#logger
logger = logging.getLogger('OCR PDF ')
fh = logging.FileHandler('./logs/ocr_pdf.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)


import ocrmypdf

from lxf.services.measure_time import measure_time


#ocrmypdf.Verbosity(0) # Suppress most messages
#ocrmypdf.configure_logging(verbosity=0, progress_bar_friendly= False)



@measure_time
def do_ocr(input_pdf, output_pdf)->ocrmypdf.ExitCode:
    # if (maxPagesLimit>0) :
    try :    
        custom_config = r'--oem 1 --oem 3 --psm 6 ' 
        ocr_result = ocrmypdf.ocr(input_pdf,
                            output_pdf,
                            output_type='pdf',
                            skip_text=True,
                            force_ocr=False,
                            tesseract_oem=3, #Tesseract + LSTM ; 0 - original Tesseract only; 1 - neural nets  LSTM  only;  2  -Tesseract + LSTM; 3 - default
                            tesseract_pagesegmode=6, # Assume a single uniform block of text.
                            keep_temporary_files=False,
                            clean=True,language="fra+eng",
                            invalidate_digital_signatures=True,
                            use_threads=True,
                            progress_bar=False,
                            deskew=False,
                            author="Python librairie PyLexfluent",
                            keywords="OCR")
        return ocr_result
    except Exception as ex:
        logger.error(f"Exception pendant l'ocerisation de {input_pdf}")
        return ocrmypdf.ExitCode.other_error


@measure_time
def do_ocr_directory(directory_source:str,directory_dest:str):
    for root, dirs, files in os.walk(directory_source):
        if files!=[] :
            category='_'.join(root.split('/')[-3:]).replace(' ','-')
            logger.debug(f"Category: {category}")
            for f in files :  
                if '.pdf' in f:   
                    input_pdf = root+'/'+f
                    output_ocred_pdf = directory_dest+"/"+category+'#'+f 
                    # if file exists in ocr directory skip it 
                    if os.path.exists(output_ocred_pdf) == False : 
                        result = do_ocr(input_pdf,output_ocred_pdf)
                        #logger.debug(f"OCR Result for {f}: {result} ")





import logging
from lxf.settings import get_logging_level

#logger
logger = logging.getLogger('Keys words')
fh = logging.FileHandler('./logs/keysword.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)


from lxf.domain.keyWord import KeyWord
from lxf.services.measure_time import measure_time_async
from lxf.domain.keyswordsandphrases import KeysWordsAndPhrases, KeysWordsPhrases



@measure_time_async  
async def get_keys_words(self,text:str,isSorted:bool=False,threshold:float=0.1)->KeysWordsPhrases:
    """
    Get the keys words and the keys phrases from a free text
    """
    keysWordsPhrasesHelper:KeysWordsAndPhrases = KeysWordsAndPhrases(text)
    logger.debug(f"Threshold {threshold}")
    freq_mots= keysWordsPhrasesHelper.get_key_words(isSorted=isSorted, threshold=threshold)
    # convert data to KeysWordsPhrases object 
    result:KeysWordsPhrases = KeysWordsPhrases()
    for mot in freq_mots:
        kword:KeyWord = KeyWord()
        kword.word=mot
        #logger.debug(f"Word: {mot}")
        kword.freq=freq_mots[mot]
        #logger.debug(f"Freq Word: {kword.freq}")
        result.keysWords.append(kword)
    return result   


import logging
from lxf.settings import get_logging_level, load_model
#logger
logger = logging.getLogger('Keys words and phrases')
fh = logging.FileHandler('./logs/keysword_and_phrases.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)

# Loading NLP
nlp_ts =load_model()

import math
import random
import time
import re
from typing import List
from pydantic import BaseModel,Field

from lxf.domain.keyWord import KeyWord 
from lxf.domain.keyphrase import KeyPhrase
from lxf.services.measure_time import measure_time






from langchain.text_splitter import RecursiveCharacterTextSplitter

text_splitter = RecursiveCharacterTextSplitter(
    chunk_size = 4096,
    chunk_overlap=256,
    length_function = len,
    is_separator_regex = False,
)

from string import punctuation
from spacy.lang.fr.stop_words import STOP_WORDS
chiffres ={"0","1","2","3","4","5","6","7","8","9"}
STOP_WORDS.update(chiffres)
symbols={"€","£","|","§","*"}
STOP_WORDS.update(symbols)
punctuations ={punct for punct in punctuation}
STOP_WORDS.update(punctuations)
miscellanous ={"mr","m.","mme","mlle","me","dr","aucun"}
STOP_WORDS.update(miscellanous)

@measure_time
def sanitize_text(text:str)->str :

    text=text.replace('(',' ( ').replace(')',' ) ').replace('–','-').replace('’','\'').replace("\n"," ").strip()
    # for stop in STOP_WORDS :
    #     text=text.replace(stop,"")
    # remove extra space
    regex = r" {2,}"
    subst = " "
    # You can manually specify the number of replacements by changing the 4th argument
    return re.sub(regex, subst, text, 0, re.MULTILINE | re.DOTALL)

@measure_time
def sub_divide(document,sep:str):
    result=[]
    for phrase in document:
        sub_phrases = phrase.split(sep)
        for sub in sub_phrases:
            result.append(sanitize_text(sub))
    return result

chunks_limit:int=30
class KeysWordsPhrases(BaseModel):
	keysWords:List[KeyWord]=Field(default=[])
	keysPhrases:List[KeyPhrase]=Field(default=[])
	class Config:       
		arbitrary_types_allowed = True 
		json_schema_extra = {
			"example": {
				"keysWords": [
					{
						"word": "montant",
						"freq": 1
					}
				],
				"keysPhrases": []
			}		
		}


class KeysWordsAndPhrases:
    """
    Keys words and keys phrases from Text
    """
    def __init__(self,text:str) :
        self.internalText=text[:nlp_ts.max_length-1024]
    
    @measure_time 
    def get_key_words(self,
        importantes_entites=["ORG_PUBLIC","PERSONNE","GARANTIE_FINANCIERE","QUALITE","ARTICLE_LOI","CODE_JURIDIQUE","TAUX_BANCAIRE","DELEGATION_BANQUE_ASSURANCE","METIER","VILLE","ORG"],
        importants_lemma=["testament","cadastre","signature","procuration","maritale","métier","partage","iban","vente","répertoire","banque","syndic","convention","bail","confidentialité","imposition",
                            "mairie","état-civil","officiel","authentique","copie","exécutoire","mariage","naissance","décés","emprunteur","assurance","hypothèque","caution","devis",
                            "prêteur","diagnostic","certificat","urbanisme","prêt","représenter","facture","honoraire","signer","délèguer","authentique","crédit","emprunt","relance","devoir",
                            "taux","acte","prêteur","emprunteur","héritier","honoraire","notaire","clerc","avocat","huissier","juge","relance","réquisition","certificat","minute","brevet",
                            "donataire","donateur","donation","statut","kbis","contrat","kbis","greffe","tribunal","commerce","trésor","impôt","ursaff","juridique","mandat"],
        threshold:float=0.15,
        isSorted:bool=True,
        entity_booster_inc:float=4,
        lemma_booster_inc:float=4):
        """
        """
        try:   
            t= time.time()                     
            logger.debug("Start measuring get_keys_words")
            text = self.internalText
            if text ==None or text == "":                
                return None
            text = sanitize_text(text)
            #logger.debug(f"Text récupéré: \n{text}")
            splitted_text=text_splitter.split_text(text)
            logger.debug(f"Splitted (Len={len(splitted_text)}) text completed at {(time.time()-t):.4f}")    

            freq_mots={}
            limit:int=len(splitted_text)
            if limit>chunks_limit:
                limit=chunks_limit                
            for chunk in splitted_text[:limit] :
                t_chunk=time.time()
                document = nlp_ts(chunk)
                logger.debug(f"NLP chunk document loaded in {(time.time()-t_chunk):.4f} secondes")
                for mot in document:   
                    # ne pas prendre pas les mots de taille inférieur ou égale 2 caractères
                    if len(mot) >2 :             
                        mot_lower= mot.lemma_.lower().replace('(','').replace(')','')
                        if not (mot.is_digit or mot.is_space or mot.is_bracket or mot.is_punct or mot.is_quote or mot.like_num) and len(mot_lower)>1 :
                            if mot_lower not in STOP_WORDS :
                                if mot_lower not in freq_mots.keys():
                                    freq_mots[mot_lower]=1
                                else :
                                    freq_mots[mot_lower]+=1
                                # boost mot 
                                if mot_lower in importants_lemma:
                                    freq_mots[mot_lower]+= lemma_booster_inc                                
            logger.debug(f"Freq_mots step 1 completed at {(time.time()-t):.4f}")
            if freq_mots !={}  :
                plus_grande_freq=max(freq_mots.values())
                for mot in freq_mots:
                    freq_mots[mot]/=plus_grande_freq
            else: 
                logger.warning("Freq_mots est null")
                # renvoyer dictionnaire vide
                return {}

            logger.debug(f"Freq_mots completed at {(time.time()-t):.4f}")
                            
            # applied threshold on keywords
            # remove all items under threshold value 
            keysWords={}
            logger.debug(f"Threshold : {threshold}")
            if freq_mots!=None :
                for mot,freq in freq_mots.items():
                    if freq >= threshold : 
                        keysWords[mot]=freq  

            logger.debug(f"freq_mot threshold completed at {(time.time()-t):.4f}") 
            # Cleanup 
            if freq_mots!=None :
                freq_mots.clear()   
            if isSorted and keysWords.items!=None:
                freq_mot_sorted= dict(sorted(keysWords.items(),key=lambda x:x[1],reverse=True))
                logger.debug(f"freq_mot sorted at {(time.time()-t):.4f}")
                return freq_mot_sorted
            return keysWords
        except Exception as ex :
            logger.critical(f"Exception KeysWordsAndPhrases.get_keys_words:  {ex}")
            return {}

    @measure_time
    def get_key_words_phrases(self,
                        importantes_entites=["ORG_PUBLIC","PERSONNE","GARANTIE_FINANCIERE","QUALITE","ARTICLE_LOI","CODE_JURIDIQUE","TAUX_BANCAIRE","DELEGATION_BANQUE_ASSURANCE","METIER","VILLE","ORG"],
                        importants_lemma=["vente","prêt","représenter","facture","honoraire","signer","délèguer","procuration","authentique","crédit","emprunt","solde","montant","relance","devoir","taux","acte","avoir","prêteur","emprunteur","héritier","honoraire","notaire","avocat","huissier","juge","relance","donataire","donateur","donation","achat"],
                        threshold:float=0.1,
                        isSorted:bool=True,
                        entity_booster_inc:float=4.5,
                        lemma_booster_inc:float=2.5):
        try:                        
            text = self.internalText
            if text ==None or text == "":                
                raise ValueError("aucun texte extrait")
            #logger.debug(f"Text récupéré: \n{text}")
            t = time.time()
            splitted_text=text_splitter.split_text(text)
            logger.debug(f"Splitted (Len={len(splitted_text)}) text completed at {(time.time()-t):.4f}") 
            
            freq_mots={}
            limit:int=len(splitted_text)
            if limit>30:
                limit=30
            for chunk in splitted_text[:limit] :
                t_chunk=time.time()
                document = nlp_ts(chunk)
                logger.debug(f"NLP chunk document loaded in {(time.time()-t_chunk):.4f} secondes")            
                for mot in document:                
                    mot_lower= mot.lemma_.lower()
                    if not (mot.is_digit or mot.is_space or mot.is_bracket or mot.is_punct or mot.is_quote or mot.like_num) and len(mot_lower)>1 :

                        if mot_lower not in STOP_WORDS :
                            if mot_lower not in freq_mots.keys():
                                freq_mots[mot_lower]=1
                            else :
                                freq_mots[mot_lower]+=1
                            # boost mot 
                            if mot_lower in importants_lemma:
                                freq_mots[mot_lower]+= lemma_booster_inc
                                logger.debug(f"Boost mot by LEMMA: {mot_lower} freq : {freq_mots[mot_lower]}")

            plus_grande_freq=max(freq_mots.values())

            for mot in freq_mots:
                freq_mots[mot]/=plus_grande_freq
           
            # Phrases scoring 
            phrases=sub_divide([text],sep=".")
            phrases=sub_divide(phrases,sep=";") 

            phrases=sub_divide(phrases,sep="- ")             
            scores_phrases={}
            for phrase in phrases:
                ph=nlp_ts(phrase)
                nb_mots_comptes=0
                for mot in ph:              
                    if not (mot.is_digit or mot.is_space or mot.is_bracket or mot.is_punct or mot.is_quote or mot.like_num) and len(mot)>1: 
                        lower_mot=mot.lemma_.lower()  
                        if not lower_mot in STOP_WORDS :
                            if lower_mot in freq_mots.keys() :
                                nb_mots_comptes+=1
                                if ph in scores_phrases.keys():
                                    scores_phrases[ph]+=freq_mots[lower_mot]
                                else :
                                    scores_phrases[ph]=freq_mots[lower_mot]
                #normalized by the number of words taken by the phrase score computation
                if nb_mots_comptes > 0 :
                    scores_phrases[ph]= float(scores_phrases[ph])/nb_mots_comptes
            # boosting the phrases scoring based on entity type or lemma
            maxScore:float=0.0
            for phrase, score in scores_phrases.items():
                entity_booster=0
                lemma_booster=0
                for ent in phrase.ents :
                    logger.debug(f"Entité {ent.label_} => {ent.text}")
                    if ent.label_ in importantes_entites:
                        entity_booster += entity_booster_inc
                for token in phrase:
                    if token.lemma_.lower() in importants_lemma:
                        lemma_booster += lemma_booster_inc
                bias = random.random()
                booster = math.log(entity_booster+lemma_booster+bias)
                if booster > 0 : 
                    # if no entity found or lemma , because log(x)<0 when x <1  where x=entity_booster+lemma_booster+bias 
                    scores_phrases[phrase]=float(score)+booster   
                if scores_phrases[phrase]> maxScore :
                    maxScore = scores_phrases[phrase]
            if maxScore>0 :
                # normalized
                for phrase, score in scores_phrases.items():
                    scores_phrases[phrase]=float(score)/maxScore                     
            # applied threshold on keywords
            # remove all items under threshold value 
            keysWords={}
            logger.debug(f"Threshold : {threshold}")
            for mot,freq in freq_mots.items():
                if freq >= threshold : 
                    keysWords[mot]=freq  
   
            # Cleanup 
            freq_mots.clear()   
            if isSorted :
                return dict(sorted(keysWords.items(),key=lambda x:x[1],reverse=True)), scores_phrases # don't sort the phrase cause meaning will be losted
            return keysWords, scores_phrases
        except Exception as ex :
            raise Exception(f"Exception KeysWordsAndPhrases.get_keys_words:  {ex}")



import os
import logging

from lxf.settings import get_logging_level, enable_tqdm

###################################################################

logger = logging.getLogger('cni ocr from PDF')
fh = logging.FileHandler('./logs/cni_ocr_from_pdf.log')
fh.setLevel(get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(get_logging_level())
logger.addHandler(fh)
#################################################################
from pathlib import Path

from pdf2image import convert_from_path
from tqdm import tqdm

from lxf.services.measure_time import measure_time_async

from typing import List

from lxf.domain.cni import Cni
from lxf.ai.ocr.cni_utilities.recognize_zones import recognize

path_images_cni = "./data/images/temp/images-cni"

def check_path() ->None:
    """
    """
    data_path =Path("./data")
    if not data_path.exists: data_path.mkdir()
    image_path=Path("./data/images")    
    if not image_path.exists(): image_path.mkdir()
    temp_path=Path("./data/images/temp")
    if not temp_path.exists() : temp_path.mkdir()
    temp_document_path=Path(path_images_cni)
    if not temp_document_path.exists() : temp_document_path.mkdir()


@measure_time_async
async def do_cni_ocr_from_pdf(path_filename:str)->List[Cni]:
    """
    Paramètre le nom du fichier pdf contenant les CNI possibles
    retourne toutes les CNI trouvées
    """
    check_path()
    if os.path.exists(path_filename) ==False :
        logger.error(f"Fichier introuvable ({path_filename})")
        return None
    pages = convert_from_path(path_filename)
    cni_candidates:List[Cni]=[]
    for i in tqdm(range(len(pages)),desc="Analyse des pages ",disable=not enable_tqdm):
        ## Save image to file
        image_path = f"{path_images_cni}/page_{i}.jpeg"
        pages[i].save(image_path,'JPEG')    
        cni:Cni = recognize(image_path)
        if cni!=None :
            cni_candidates.append(cni)
    return cni_candidates




from typing import List,Optional
from xml.dom.minidom import Entity
from pydantic import BaseModel,Field

class VidalGeneric(BaseModel):
    nom:str=Field(default="",description="Nomenclature généric Vidal")
    
class VidalProduc(BaseModel):
    code:str=Field(default="",description="Code Vidal")
    product:str=Field(default="",description="Nomenclature Vidal")
    generics:List[VidalGeneric]=[]
    
class Medicament(BaseModel):
    nom:str=Field(default="",description="nom du médicament")
    start_text:int=Field(default=0,description='Début du texte')
    end_text:int=Field(default=0,description='Fin du texte')
    medicament:str=Field(default="",description="Texte reconnu brute")
    posologie:str=Field(default='',description="Posologie application médicament")
    vidal_references:List[VidalProduc]=[]
    class Config:        
        arbitrary_types_allowed = True

class AnalyseOrdonnance(BaseModel):
    nom:str=Field(default='',description="Nom de l'ordonnance")
    raw_text:str=Field(default='',description="Texte océrisé de l'ordonnance")
    medicaments:List[Medicament]=[]
    class Config:        
        arbitrary_types_allowed = True


import logging
import lxf.settings as settings
settings.enable_tqdm=False
logger = logging.getLogger('Training Data')
fh = logging.FileHandler('./logs/training_data.log')
fh.setLevel(settings.get_logging_level())
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
fh.setFormatter(formatter)
logger.setLevel(settings.get_logging_level())
logger.addHandler(fh)
from lxf.domain.entity import Entity

class DocumentTrainingData(Entity):
    def __init__(self,dictionary:dict=None) :
        if dictionary==None:
            super().__init__(sid="",parent_id="")
            self.name=""
            self.model=""
            self.famille=""
            self.category=""
            self.sub_category=""
            self.key_words={}
            self.key_phrases={}
        else :
            super().__init__(sid=dictionary["sid"],parent_id=dictionary["parent_id"])
            self.name=dictionary["name"]
            self.model=dictionary["model"]
            self.famille=dictionary["famille"]
            self.category=dictionary["category"]
            self.sub_category=dictionary["sub_category"]
            self.key_words = dictionary["key_words"]
            self.key_phrases=dictionary["key_phrases"]            
        

    def __str__(self)->str:
        return f"{DocumentTrainingData.__name__}\n{Entity.__str__(self)}\nclassification: {self.famille}_{self.category}_{self.sub_category}\nkey_words:{self.key_words}\nkey_phrases:{self.key_phrases}"
 




import itertools
from sklearn.metrics import confusion_matrix
import numpy as np
import matplotlib.pyplot as plt

random_seed_base=13

# params for learning
embedding_dim=32
lr_base = 1e-3
epochs=180

# global variables
concatenate_col = True
X_col_key_phrases ='PhrasesCles'
X_col_key_words='MotsCles'
y_col='Classification'
drop_columns_for_concatenate_col=['Classification','Libelle']
drop_columns=['Classification','Libelle','MotsCles']
# Set maximum_sequence length as all samples are not of the same length.
# Vocabulary size and number of words in a sequence.
vocab_size = 10000
sequence_length = 256
if concatenate_col ==False :
  ngrams = 4
else :
  ngrams=None
batch_size=32

import tensorflow as tf


def vectorize_text(vectorize_layer,text, label):
  text = tf.expand_dims(text, -1)
  return vectorize_layer(text), label

def create_model(class_names,model_name,vectorize_layer):
    #define tensorboard callback
    tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir="logs/"+model_name)
    # Create a learning rate scheduler callback
    lr_scheduler = tf.keras.callbacks.LearningRateScheduler(lambda epoch: lr_base * 10**(epoch/epochs))
    #set seed 
    tf.random.set_seed(random_seed_base)
    #create the model
    model=tf.keras.Sequential(name=model_name,layers=[
        vectorize_layer,
        tf.kerasEmbedding(vocab_size,embedding_dim,name='embedding'),
        tf.keras.GlobalAveragePooling1D(),
        tf.kereas.Dense(embedding_dim,activation='relu'),
        tf.keras.Dense(len(class_names),activation='softmax')
    ])
    #compile the model 
    model.compile(loss=tf.keras.losses.SparseCategoricalCrossentropy(),
                optimizer=tf.keras.optimizers.Adam(),
                metrics=["accuracy"])
    return model, [tensorboard_callback,lr_scheduler]


# Note: The following confusion matrix code is a remix of Scikit-Learn's 
# plot_confusion_matrix function - https://scikit-learn.org/stable/modules/generated/sklearn.metrics.plot_confusion_matrix.html
# and Made with ML's introductory notebook - https://github.com/GokuMohandas/MadeWithML/blob/main/notebooks/08_Neural_Networks.ipynb
# Our function needs a different name to sklearn's plot_confusion_matrix
def make_confusion_matrix(y_true, y_pred, classes=None, figsize=(10, 10), text_size=20): 
  """Makes a labelled confusion matrix comparing predictions and ground truth labels.

  If classes is passed, confusion matrix will be labelled, if not, integer class values
  will be used.

  Args:
    y_true: Array of truth labels (must be same shape as y_pred).
    y_pred: Array of predicted labels (must be same shape as y_true).
    classes: Array of class labels (e.g. string form). If `None`, integer labels are used.
    figsize: Size of output figure (default=(10, 10)).
    text_size: Size of output figure text (default=15).
  
  Returns:
    A labelled confusion matrix plot comparing y_true and y_pred.

  Example usage:
    make_confusion_matrix(y_true=test_labels, # ground truth test labels
                          y_pred=y_preds, # predicted labels
                          classes=class_names, # array of class label names
                          figsize=(15, 15),
                          text_size=10)
  """  
  # Create the confustion matrix
  cm = confusion_matrix(y_true, y_pred)
  cm_norm = cm.astype("float") / cm.sum(axis=1)[:, np.newaxis] # normalize it
  

  # Plot the figure and make it pretty
  fig, ax = plt.subplots(figsize=figsize)
  cax = ax.matshow(cm, cmap=plt.cm.Greens) # colors will represent how 'correct' a class is, darker == better
  fig.colorbar(cax)

  # Are there a list of classes?
  if classes.any():
    labels = classes
    n_classes = len(labels) # find the number of classes we're dealing with
  else:
    labels = np.arange(cm.shape[0])
    n_classes = cm.shape[0] # find the number of classes we're dealing with
  
  # Label the axes
  ax.set(title="Confusion Matrix",
         xlabel="Predicted label",
         ylabel="True label",
         xticks=np.arange(n_classes), # create enough axis slots for each class
         yticks=np.arange(n_classes), 
         xticklabels=labels, # axes will labeled with class names (if they exist) or ints
         yticklabels=labels)
  
  # Make x-axis labels appear on bottom
  ax.xaxis.set_label_position("bottom")
  ax.xaxis.tick_bottom()

  # Set the threshold for different colors
  threshold = (cm.max() + cm.min()) / 2.
  print(f"Threashold : {threshold:.2f}\n")
  # Plot the text on each cell
  for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, f"{cm_norm[i, j]*100:.1f}% ({cm[i,j]})",
             horizontalalignment="center",
             color="white" if cm[i, j] >= threshold else "red",
             size=text_size)